"""pyiem.plot.util Plotting Utilities!"""
# pylint: disable=import-outside-toplevel
import os

import numpy as np
import cartopy.crs as ccrs
import matplotlib.path as mpath
import matplotlib.patches as mpatches
import matplotlib.image as mpimage
import matplotlib.colors as mpcolors
import matplotlib.patheffects as PathEffects
from pyiem import reference
from pyiem.plot.colormaps import stretch_cmap

DATADIR = os.sep.join([os.path.dirname(__file__), "..", "data"])
LOGO_BOUNDS = (0.005, 0.91, 0.08, 0.086)


def pretty_bins(minval, maxval, bins=8):
    """Return a **smooth** binning that encloses the min and max value.

    The returned array is +1 in size of the bins specified, since we want the
    bin edges.

    Args:
      minval (real): minimum value to enclose.
      maxval (real): maximum value to enclose.
      bins (int): number of bins to generate
    Returns:
      ``np.array`` of bins"""
    center = (maxval + minval) / 2.0
    return centered_bins(maxval - center, on=center, bins=bins)


def centered_bins(absmax, on=0, bins=8):
    """Return a **smooth** binning around some number.

    The returned array is +1 in size of the bins specified, since we want the
    bin edges.

    Args:
      absmax (real): positive distance from the `on` value for bins to enclose.
      on (real): where to center these bins.
      bins (int): number of bins to generate
    Returns:
      ``np.array`` of bins"""
    # We want an array returned with +1 bins
    sz = float(bins) / 2.0
    mx = (absmax) / sz
    interval = np.around(mx, 2)
    # We don't want any floating point numbers over 5
    if interval > 5 and interval * 100 % 100 != 0:
        interval = np.ceil(mx)
        return np.linspace(on - sz * interval, on + sz * interval, bins + 1)
    if mx == interval:
        return np.linspace(on - sz * interval, on + sz * interval, bins + 1)
    # Find a pretty interval >= mx
    c = [0.001, 0.005, 0.01, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.75, 1, 1.5, 5]
    for interval in c:
        if interval > mx:
            break
    return np.linspace(on - sz * interval, on + sz * interval, bins + 1)


def draw_logo(fig, filename):
    """Place the logo."""
    fn = "%s/%s" % (DATADIR, filename)
    if not os.path.isfile(fn):
        return
    # Create a fake axes to place this Logo
    logo = mpimage.imread(fn)
    # imshow messes with the aspect, so about the best we can do here is
    # pin it to the upper edge
    ax = fig.add_axes(
        LOGO_BOUNDS, frameon=False, yticks=[], xticks=[], anchor="NW"
    )
    ax.imshow(logo, aspect="equal", zorder=-1)


def fontscale(ratio, fig=None):
    """Return a font size suitable for this NDC ratio.

    Args:
      ratio (float): value between 0 and 1
      fig (matplotlib.Figure,optional): The Figure of interest

    Returns:
      float: font size
    """
    if fig is None:
        # Lazy import to prevent default backend setting
        import matplotlib.pyplot as plt

        fig = plt.gcf()
    bbox = fig.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    return bbox.height * fig.dpi * ratio


def fitbox(fig, text, x0, x1, y0, y1, **kwargs):
    """Fit text into a NDC box.

    Args:
      textsize (int, optional): First attempt this textsize to see if it fits.
    """
    figbox = fig.get_window_extent().transformed(
        fig.dpi_scale_trans.inverted()
    )
    # need some slop for decimal comparison below
    px0 = x0 * fig.dpi * figbox.width - 0.15
    px1 = x1 * fig.dpi * figbox.width + 0.15
    py0 = y0 * fig.dpi * figbox.height - 0.15
    py1 = y1 * fig.dpi * figbox.height + 0.15
    xanchor = x0
    if kwargs.get("ha", "") == "center":
        xanchor = x0 + (x1 - x0) / 2.0
    yanchor = y0
    if kwargs.get("va", "") == "center":
        yanchor = y0 + (y1 - y0) / 2.0
    txt = fig.text(
        xanchor,
        yanchor,
        text,
        fontsize=kwargs.get("textsize", 50),
        ha=kwargs.get("ha", "left"),
        va=kwargs.get("va", "bottom"),
        color=kwargs.get("color", "k"),
    )

    def _fits(txt):
        """Test for fitting."""
        tb = txt.get_window_extent(fig.canvas.get_renderer())
        return tb.x0 >= px0 and tb.x1 < px1 and tb.y0 >= py0 and tb.y1 <= py1

    if not _fits(txt):
        for size in range(50, 1, -2):
            txt.set_fontsize(size)
            if _fits(txt):
                break
    return txt


def make_axes(ndc_axbounds, geoextent, projection, aspect):
    """Factory for making an axis

    Args:
      ndc_axbounds (list): the NDC coordinates of axes to create
      geoextent (list): x0,x1,y0,y1 the lon/lon extent of the axes to create
      projection (ccrs.Projection): the projection of the axes
      aspect (str): matplotlib's aspect of axes

    Returns:
      ax
    """
    # Lazy import to prevent backend setting
    import matplotlib.pyplot as plt

    ax = plt.axes(
        ndc_axbounds,
        projection=projection,
        aspect=aspect,
        facecolor=(0.4471, 0.6235, 0.8117),
    )
    # Get the frame at the proper zorder
    for _k, spine in ax.spines.items():
        spine.set_zorder(reference.Z_FRAME)
    ax.set_extent(geoextent)
    if aspect != "equal":
        return ax
    # Render the canvas so we know what happened with our axis
    ax.figure.canvas.draw()
    # Compute the current NDC extent of the axes
    ndc_bbox = ax.get_position()
    # pixel_bbox = ax.get_window_extent()
    (projx0, projx1, projy0, projy1) = ax.get_extent()
    # Figure out which axis got shrunk
    xscaled = ndc_bbox.width / float(ndc_axbounds[2])
    yscaled = ndc_bbox.height / float(ndc_axbounds[3])
    # compute the dx/dy of this image
    # dx = (projx1 - projx0) / pixel_bbox.width
    # dx = (projy1 - projy0) / pixel_bbox.height
    # expand one way or another to fit, via set_extent
    xneeded = (projx1 - projx0) / xscaled - (projx1 - projx0)
    yneeded = (projy1 - projy0) / yscaled - (projy1 - projy0)
    newbounds = [
        projx0 - xneeded / 2.0,
        projx1 + xneeded / 2.0,
        projy0 - yneeded / 2.0,
        projy1 + yneeded / 2.0,
    ]
    ax.set_extent(newbounds, crs=projection)
    # Render the canvas so we know what happened with our axis
    ax.figure.canvas.draw()
    return ax


def sector_setter(mp, axbounds, **kwargs):
    """use kwargs to set the MapPlot sector."""
    aspect = kwargs.get("aspect", "equal")

    if mp.sector == "cwa":
        mp.cwa = kwargs.get("cwa", "DMX")
        mp.ax = make_axes(
            axbounds,
            [
                reference.wfo_bounds[mp.cwa][0],
                reference.wfo_bounds[mp.cwa][2],
                reference.wfo_bounds[mp.cwa][1],
                reference.wfo_bounds[mp.cwa][3],
            ],
            ccrs.Mercator(),
            aspect,
        )
        mp.axes.append(mp.ax)
    elif mp.sector == "state":
        mp.state = kwargs.get("state", "IA")
        # We hard code aspect as Alaska does funny things with 'equal' set
        mp.ax = make_axes(
            axbounds,
            [
                reference.state_bounds[mp.state][0],
                reference.state_bounds[mp.state][2],
                reference.state_bounds[mp.state][1],
                reference.state_bounds[mp.state][3],
            ],
            ccrs.Mercator(),
            aspect if mp.state != "AK" else "auto",
        )
        mp.axes.append(mp.ax)
    elif mp.sector in reference.SECTORS:
        mp.ax = make_axes(
            axbounds, reference.SECTORS[mp.sector], ccrs.Mercator(), aspect
        )
        mp.axes.append(mp.ax)
    elif mp.sector == "iowawfo":
        mp.ax = make_axes(
            axbounds, [-99.6, -89.0, 39.8, 45.5], ccrs.Mercator(), aspect
        )
        mp.axes.append(mp.ax)
    elif mp.sector == "custom":
        mp.ax = make_axes(
            axbounds,
            [kwargs["west"], kwargs["east"], kwargs["south"], kwargs["north"]],
            kwargs.get("projection", ccrs.Mercator()),
            aspect,
        )
        mp.axes.append(mp.ax)

    elif mp.sector == "north_america":
        mp.ax = make_axes(
            axbounds,
            [-145.5, -2.566, 1, 46.352],
            ccrs.LambertConformal(
                central_longitude=-107.0, central_latitude=50.0
            ),
            aspect,
        )
        mp.axes.append(mp.ax)

    elif mp.sector in ["conus", "nws"]:
        mp.ax = make_axes(
            axbounds,
            [
                reference.CONUS_WEST + 14,
                reference.CONUS_EAST - 12,
                reference.CONUS_SOUTH,
                reference.CONUS_NORTH + 1,
            ],
            reference.EPSG[5070],
            aspect,
        )
        mp.axes.append(mp.ax)

        if mp.sector == "nws":
            # Create PR, AK, and HI sectors
            mp.pr_ax = make_axes(
                [0.78, 0.055, 0.125, 0.1],
                [-68.0, -65.0, 17.5, 18.6],
                ccrs.PlateCarree(central_longitude=-105.0),
                aspect,
            )
            mp.axes.append(mp.pr_ax)
            # Create AK
            mp.ak_ax = make_axes(
                [0.015, 0.055, 0.25, 0.2],
                [-179.5, -129.0, 51.08, 72.1],
                ccrs.PlateCarree(central_longitude=-105.0),
                aspect,
            )
            mp.axes.append(mp.ak_ax)
            # Create HI
            mp.hi_ax = make_axes(
                [0.47, 0.055, 0.2, 0.1],
                [-161.0, -154.0, 18.5, 22.5],
                ccrs.PlateCarree(central_longitude=-105.0),
                aspect,
            )
            mp.axes.append(mp.hi_ax)


def mask_outside_polygon(poly_verts, ax=None):
    """
    We produce a polygon that lies between the plot border and some interior
    polygon.

    POLY_VERTS is in CCW order, as this is the interior of the polygon
    """
    if ax is None:
        # Lazy import to prevent default backend setting
        import matplotlib.pyplot as plt

        ax = plt.gca()

    # Get current plot limits
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    # Verticies of the plot boundaries in clockwise order
    bound_verts = np.array(
        [
            (xlim[0], ylim[0]),
            (xlim[0], ylim[1]),
            (xlim[1], ylim[1]),
            (xlim[1], ylim[0]),
            (xlim[0], ylim[0]),
        ]
    )

    # A series of codes (1 and 2) to tell matplotlib whether to draw a lineor
    # move the "pen" (So that there's no connecting line)
    bound_codes = [mpath.Path.MOVETO] + (len(bound_verts) - 1) * [
        mpath.Path.LINETO
    ]
    poly_codes = [mpath.Path.MOVETO] + (len(poly_verts) - 1) * [
        mpath.Path.LINETO
    ]

    # Plot the masking patch
    path = mpath.Path(
        np.concatenate([bound_verts, poly_verts]), bound_codes + poly_codes
    )
    # remove data
    patch = mpatches.PathPatch(
        path, facecolor="white", edgecolor="none", zorder=reference.Z_CLIP
    )
    patch = ax.add_patch(patch)

    # Reset the plot limits to their original extents
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    return patch


def polygon_fill(mymap, geo_provider, data, **kwargs):
    """Generalized function for overlaying filled polygons on the map

    Args:
      mymap (MapPlot): The MapPlot instance
      geo_provider (dict): The dictionary of keys and geometries
      data (dict): The dictionary of keys and values used for picking colors
      **kwargs (Optional): Other things needed for mapping
        ilabel (Optional[bool]): should values be labelled? Defaults to `False`
        plotmissing (bool): should geometries not included in the `data`
        be mapped? Defaults to `True`
    """
    bins = kwargs.get("bins", np.arange(0, 101, 10))
    cmap = stretch_cmap(kwargs.get("cmap"), bins, extend=kwargs.get("extend"))
    ilabel = kwargs.get("ilabel", False)
    norm = mpcolors.BoundaryNorm(bins, cmap.N)
    lblformat = kwargs.get("lblformat", "%s")
    labels = kwargs.get("labels", dict())
    plotmissing = kwargs.get("plotmissing", True)
    for polykey, polydict in geo_provider.items():
        # our dictionary is bytes so we need str
        val = data.get(polykey, None)
        if val is None:
            if not plotmissing:
                continue
            lbl = labels.get(polykey, "-")
            c = "white"
        else:
            lbl = labels.get(polykey, lblformat % (val,))
            c = cmap(norm([val]))[0]
        # in python3, our dict types are byte arrays
        for polyi, polygon in enumerate(polydict.get("geom", [])):
            if polygon.exterior is None:
                continue
            a = np.asarray(polygon.exterior)
            for ax in mymap.axes:
                points = ax.projection.transform_points(
                    ccrs.Geodetic(), a[:, 0], a[:, 1]
                )
                p = mpatches.Polygon(
                    points[:, :2],
                    fc=c,
                    ec="k",
                    zorder=reference.Z_FILL,
                    lw=0.1,
                )
                ax.add_patch(p)
                if ilabel and polyi == 0:
                    txt = ax.text(
                        polydict.get("lon", polygon.centroid.x),
                        polydict.get("lat", polygon.centroid.y),
                        lbl,
                        zorder=100,
                        clip_on=True,
                        ha="center",
                        va="center",
                        transform=ccrs.PlateCarree(),
                    )
                    txt.set_path_effects(
                        [PathEffects.withStroke(linewidth=2, foreground="w")]
                    )

    kwargs.pop("cmap", None)
    kwargs.pop("bins", None)
    mymap.draw_colorbar(bins, cmap, norm, **kwargs)


def mask_outside_geom(ax, geom):
    """Create a white patch over the plot for what we want to ask out

    Args:
      ax (axes):
      geom (geometry):
    """
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    # Verticies of the plot boundaries in clockwise order
    verts = np.array(
        [
            (xlim[0], ylim[0]),
            (xlim[0], ylim[1]),
            (xlim[1], ylim[1]),
            (xlim[1], ylim[0]),
            (xlim[0], ylim[0]),
        ]
    )
    codes = [mpath.Path.MOVETO] + (len(verts) - 1) * [mpath.Path.LINETO]
    for geo in geom:
        ccw = np.asarray(geo.exterior)[::-1]
        points = ax.projection.transform_points(
            ccrs.Geodetic(), ccw[:, 0], ccw[:, 1]
        )
        verts = np.concatenate([verts, points[:, :2]])
        codes = np.concatenate(
            [
                codes,
                [mpath.Path.MOVETO]
                + (points.shape[0] - 1) * [mpath.Path.LINETO],
            ]
        )

    path = mpath.Path(verts, codes)
    # Removes any external data
    patch = mpatches.PathPatch(
        path, facecolor="white", edgecolor="none", zorder=reference.Z_CLIP
    )
    ax.add_patch(patch)
    # Then gives a nice semitransparent look
    patch = mpatches.PathPatch(
        path,
        facecolor="black",
        edgecolor="none",
        zorder=reference.Z_CLIP2,
        alpha=0.65,
    )
    ax.add_patch(patch)



"""A utility to load matplotlib and set the backend to AGG

Example:
   from pyiem.plot.use_agg import plt
"""
# pylint: disable=unused-import,wrong-import-position
import os

from pandas.plotting import register_matplotlib_converters
import matplotlib

matplotlib.use("agg")
import matplotlib.pyplot as plt  # noqa

# Workaround a pandas dataframe to matplotlib issue
register_matplotlib_converters()

# work around warning coming from pooch
if "TEST_DATA_DIR" not in os.environ:
    os.environ["TEST_DATA_DIR"] = "/tmp"



"""Parser and object storage of information within NWS CLI Product.
"""
import re
import datetime

from pyiem.reference import TRACE_VALUE
from pyiem.nws.product import TextProduct
from pyiem.util import LOG
from pyiem.observation import Observation
from pyiem.exceptions import CLIException

HEADLINE_RE = re.compile(
    (
        r"\.\.\.THE ([A-Z_\.\-\(\)\/\,\s]+) "
        r"CLIMATE SUMMARY FOR\s+"
        r"([A-Z]+\s[0-9]+\s+[0-9]{4})( CORRECTION)?\.\.\."
    )
)
WIND_RE = re.compile(
    r"(HIGHEST|AVERAGE|RESULTANT)\s(WIND|GUST)\s(SPEED|DIRECTION)"
)

REGIMES = [
    "WEATHER ITEM   OBSERVED TIME   RECORD YEAR NORMAL DEPARTURE LAST",
    "WEATHER ITEM   OBSERVED TIME   NORMAL DEPARTURE LAST",
    "WEATHER ITEM   OBSERVED TIME    RECORD YEAR NORMAL DEPARTURE LAST",
    "WEATHER ITEM   OBSERVED RECORD YEAR NORMAL DEPARTURE LAST",
    "WEATHER ITEM   OBSERVED TIME   RECORD YEAR",
    "WEATHER ITEM   OBSERVED TIME   RECORD YEAR NORMAL DEPARTURE",
    "WEATHER ITEM   OBSERVED RECORD YEAR NORMAL DEPARTURE",
    "WEATHER ITEM   OBSERVED",
    "WEATHER ITEM   OBSERVED RECORD YEAR NORMAL",
    "WEATHER ITEM   OBSERVED TIME   RECORD YEAR NORMAL  LAST",
    "WEATHER ITEM   OBSERVED TIME       LAST",
    "WEATHER ITEM   OBSERVED NORMAL DEPARTURE LAST",
    "WEATHER ITEM   OBSERVED TIME   NORMAL  LAST",
    "WEATHER ITEM   OBSERVED TIME   RECORD YEAR     LAST",
    "WEATHER ITEM   OBSERVED TIME",
    "WEATHER ITEM   OBSERVED TIME   NORMAL DEPARTURE",
    "WEATHER ITEM   OBSERVED NORMAL DEPARTURE",
    "WEATHER ITEM   OBSERVED TIME   RECORD NORMAL DEPARTURE LAST",
]
# label, value, time, record, year, normal, departure, last
COLS = [
    [16, 23, 30, 37, 42, 49, 56, 65],
    [16, 23, 30, None, None, 37, 44, 53],
    [16, 22, 31, 37, 43, 50, 58, 65],
    [16, 23, None, 30, 35, 42, 49, 58],
    [16, 23, 25, 37, 42, None, None, None],
    [16, 23, 30, 37, 42, 49, 56, None],
    [16, 23, None, 30, 35, 42, 49, None],
    [16, 23, None, None, None, None, None, None],
    [16, 23, None, 30, 37, None, None, None],
    [16, 23, 30, 37, 42, 49, None, 57],
    [16, 23, 30, None, None, None, None, 39],
    [16, 23, None, None, None, 30, 37, 46],
    [16, 23, 30, None, None, 37, None, 45],
    [16, 23, 30, 37, 42, None, None, 51],
    [16, 23, 30, None, None, None, None, None],
    [16, 23, 30, None, None, 37, 44, None],
    [16, 23, None, None, None, 30, 37, None],
    [16, 23, 30, 37, None, 44, 51, 60],
]
# Allow manual provision of IDS
HARDCODED = {}


def update_iemaccess(txn, entry):
    """Update the IEM Access Database."""
    if entry["access_network"] is None:
        return
    ob = Observation(
        entry["access_station"], entry["access_network"], entry["cli_valid"]
    )
    ob.load(txn)
    current = ob.data
    data = entry["data"]
    logmsg = []
    if data.get("temperature_maximum") is not None:
        climax = int(data["temperature_maximum"])
        if climax != current["max_tmpf"]:
            logmsg.append("MaxT O:%s N:%s" % (current["max_tmpf"], climax))
            current["max_tmpf"] = climax

    if data.get("temperature_minimum") is not None:
        climin = int(data["temperature_minimum"])
        if climin != current["min_tmpf"]:
            logmsg.append("MinT O:%s N:%s" % (current["min_tmpf"], climin))
            current["min_tmpf"] = climin

    if data.get("precip_month") is not None:
        val = data["precip_month"]
        if val != current["pmonth"]:
            logmsg.append("PMonth O:%s N:%s" % (current["pmonth"], val))
            current["pmonth"] = val

    if data.get("precip_today") is not None:
        val = data["precip_today"]
        if val != current["pday"]:
            logmsg.append("PDay O:%s N:%s" % (current["pday"], val))
            current["pday"] = val

    if data.get("snow_today") is not None:
        val = data["snow_today"]
        if current["snow"] is None or val != current["snow"]:
            logmsg.append("Snow O:%s N:%s" % (current["snow"], val))
            current["snow"] = val

    if not logmsg:
        return
    ob.save(txn, skip_current=True)
    LOG.info(
        "%s (%s) %s",
        entry["access_station"],
        entry["cli_valid"].strftime("%y%m%d"),
        ",".join(logmsg),
    )


def trace_r(val):
    """ Convert our value back into meaningful string """
    if val is None:
        return "Missing"
    if val == TRACE_VALUE:
        return "Trace"
    return val


def get_number(text):
    """ Convert a string into a number, preferable a float! """
    if text is None:
        return None
    text = text.strip()
    if text == "":
        retval = None
    elif text == "MM":
        retval = None
    elif text == "T":
        retval = TRACE_VALUE
    else:
        number = re.findall(r"[\-\+]?\d*\.\d+|[\-\+]?\d+", text)
        if len(number) == 1:
            if text.find(".") > 0:
                retval = float(number[0])
            else:
                retval = int(number[0])
        else:
            LOG.info("get_number() failed for |%s|", text)
            retval = None
    return retval


def convert_key(text):
    """ Convert a key value to something we store """
    if text is None:
        return None
    if text == "YESTERDAY":
        return "today"
    if text == "TODAY":
        return "today"
    if text == "MONTH TO DATE":
        return "month"
    if text.startswith("SINCE "):
        return text.replace("SINCE ", "").replace(" ", "").lower()
    LOG.info("convert_key() failed for |%s|", text)
    return "fail"


def make_tokens(regime, line):
    """ Turn a line into tokens based on a regime """
    mycols = COLS[regime]
    tokens = []
    pos = 0
    for e in mycols:
        if e is None:
            tokens.append(None)
            continue
        tokens.append(
            line[pos:e].strip() if line[pos:e].strip() != "" else None
        )
        pos = e
    for i, token in enumerate(tokens):
        if token is not None and token.startswith("R "):
            tokens[i] = token.replace("R ", "")
    return tokens


def parse_snowfall(regime, lines, data):
    """Parse the snowfall data"""
    for linenum, line in enumerate(lines):
        # skipme
        if len(line.strip()) < 14:
            continue
        tokens = make_tokens(regime, line)
        key = tokens[0].strip()
        if key == "SNOW DEPTH":
            continue
        key = convert_key(key)
        data["snow_%s" % (key,)] = get_number(tokens[1])
        data["snow_%s_record" % (key,)] = get_number(tokens[3])
        yeartest = get_number(tokens[4])
        if yeartest is not None:
            data["snow_%s_record_years" % (key,)] = [yeartest]
        data["snow_%s_normal" % (key,)] = get_number(tokens[5])
        data["snow_%s_departure" % (key,)] = get_number(tokens[6])
        data["snow_%s_last" % (key,)] = get_number(tokens[7])
        if (
            key == "today"
            and yeartest is not None
            and data["snow_%s_record_years" % (key,)][0] is not None
        ):
            while (linenum + 1) < len(lines) and len(
                lines[linenum + 1].strip()
            ) == 4:
                data["snow_today_record_years"].append(int(lines[linenum + 1]))
                linenum += 1


def parse_precipitation(regime, lines, data):
    """ Parse the precipitation data """
    for linenum, line in enumerate(lines):
        if len(line.strip()) < 20:
            continue
        tokens = make_tokens(regime, line)
        key = convert_key(tokens[0])
        if key is None:
            continue

        data["precip_%s" % (key,)] = get_number(tokens[1])
        data["precip_%s_record" % (key,)] = get_number(tokens[3])
        yeartest = get_number(tokens[4])
        if yeartest is not None:
            data["precip_%s_record_years" % (key,)] = [yeartest]
        data["precip_%s_normal" % (key,)] = get_number(tokens[5])
        data["precip_%s_departure" % (key,)] = get_number(tokens[6])
        data["precip_%s_last" % (key,)] = get_number(tokens[7])
        if (
            key == "today"
            and yeartest is not None
            and data["precip_%s_record_years" % (key,)][0] is not None
        ):
            while (linenum + 1) < len(lines) and len(
                lines[linenum + 1].strip()
            ) == 4:
                data["precip_today_record_years"].append(
                    int(lines[linenum + 1])
                )
                linenum += 1


def parse_temperature(regime, lines, data):
    """Here we parse a temperature section"""
    for linenum, line in enumerate(lines):
        if len(line.strip()) < 18:
            continue
        tokens = make_tokens(regime, line)
        key = tokens[0].strip().lower()
        if key.upper() not in ["MAXIMUM", "MINIMUM", "AVERAGE"]:
            continue
        data["temperature_%s" % (key,)] = get_number(tokens[1])
        if tokens[2] is not None:
            data["temperature_%s_time" % (key,)] = tokens[2]
        if tokens[3] is not None:
            data["temperature_%s_record" % (key,)] = get_number(tokens[3])
        if tokens[4] is not None:
            n = get_number(tokens[4])
            if n is not None:
                data["temperature_%s_record_years" % (key,)] = [n]
        if tokens[5] is not None:
            data["temperature_%s_normal" % (key,)] = get_number(tokens[5])
            # Check next line(s) for more years
            while (linenum + 1) < len(lines) and len(
                lines[linenum + 1].strip()
            ) == 4:
                data["temperature_%s_record_years" % (key,)].append(
                    int(lines[linenum + 1])
                )
                linenum += 1


def parse_sky_coverage(lines, data):
    """Turn section into data."""
    asc = "AVERAGE SKY COVER"
    for line in lines:
        if line.strip().startswith(asc):
            try:
                data["average_sky_cover"] = float(line.replace(asc, ""))
            except ValueError:
                pass


def parse_wind(lines, data):
    """Parse any wind information."""
    # hold your nose here
    # make everything space seperated
    content = " ".join((" ".join(lines[1:])).strip().split())
    tokens = WIND_RE.findall(content)
    for token in tokens:
        content = content.replace(" ".join(token), ";")
    vals = content[1:].split(";")
    for token, val in zip(tokens, vals):
        data[("_".join(token)).lower()] = get_number(val)


def _compute_station_ids(prod, cli_station, is_multi):
    """Figure out what the various station IDs are."""
    # Can't always use the AFOS as the station ID :(
    if is_multi:
        station = None
        for st in prod.nwsli_provider:
            if prod.nwsli_provider[st]["name"].upper() == cli_station:
                station = st[1:]  # drop first char
                break
        if station is None:
            raise CLIException(
                "Unknown CLI Station Text: |%s|" % (cli_station,)
            )
    else:
        station = prod.afos[3:]
    db_station = "%s%s" % (prod.source[0], station)
    db_station = HARDCODED.get(db_station, db_station)
    access_station = db_station if not db_station.startswith("K") else station
    # Compute the Network
    if access_station not in prod.nwsli_provider:
        prod.warnings.append(
            "Unknown IEMAccess station |%s|" % (access_station,)
        )
        network = None
    else:
        network = prod.nwsli_provider[access_station]["network"]

    return db_station, access_station, network


class CLIProduct(TextProduct):
    """
    Represents a CLI Daily Climate Report Product
    """

    def __init__(
        self, text, utcnow=None, ugc_provider=None, nwsli_provider=None
    ):
        """ constructor """
        TextProduct.__init__(self, text, utcnow, ugc_provider, nwsli_provider)
        # Hold our parsing results as an array of dicts
        self.data = []
        self.regime = None
        # Sometimes, we get products that are not really in CLI format but
        # are RER (record event reports) with a CLI AWIPS ID
        if self.wmo[:2] != "CD":
            LOG.info(
                "Product %s skipped due to wrong header", self.get_product_id()
            )
            return
        sections = self.find_sections()
        for section in sections:
            # We have meat!
            self.compute_diction(section)
            entry = {}
            entry["cli_valid"], entry["cli_station"] = self.parse_cli_headline(
                section
            )
            (
                entry["db_station"],
                entry["access_station"],
                entry["access_network"],
            ) = _compute_station_ids(
                self, entry["cli_station"], len(sections) > 1
            )
            entry["data"] = self.parse_data(section)
            self.data.append(entry)

    def find_sections(self):
        """Some trickery to figure out if we have multiple reports

        Returns:
          list of text sections
        """
        sections = []
        for section in self.unixtext.split("&&"):
            if not HEADLINE_RE.findall(section.replace("\n", " ")):
                continue
            tokens = re.findall("^WEATHER ITEM.*$", section, re.M)
            if not tokens:
                raise CLIException("Could not find 'WEATHER ITEM' within text")
            if len(tokens) == 1:
                sections.append(section)
                continue
            # Uh oh, we need to do some manual splitting
            pos = []
            for match in re.finditer(HEADLINE_RE, section.replace("\n", " ")):
                pos.append(match.start())
            pos.append(len(section))
            for i, p in enumerate(pos[:-1]):
                sections.append(section[max([0, p - 10]) : pos[i + 1]])
        return sections

    def compute_diction(self, text):
        """ Try to determine what we have for a format """
        tokens = re.findall("^WEATHER ITEM.*$", text, re.M)
        diction = tokens[0].strip()
        if diction not in REGIMES:
            raise CLIException(
                "Unknown diction found in 'WEATHER ITEM'\n|%s|" % (diction,)
            )

        self.regime = REGIMES.index(diction)

    def get_jabbers(self, uri, _=None):
        """ Override the jabber message formatter """
        url = "%s?pid=%s" % (uri, self.get_product_id())
        res = []
        for data in self.data:
            mess = (
                "%s %s Climate Report: High: %s Low: %s "
                "Precip: %s Snow: %s %s"
            ) % (
                data["cli_station"],
                data["cli_valid"].strftime("%b %-d"),
                data["data"].get("temperature_maximum", "M"),
                data["data"].get("temperature_minimum", "M"),
                trace_r(data["data"].get("precip_today", "M")),
                trace_r(data["data"].get("snow_today", "M")),
                url,
            )
            htmlmess = (
                '%s <a href="%s">%s Climate Report</a>: High: %s '
                "Low: %s Precip: %s Snow: %s"
            ) % (
                data["cli_station"],
                url,
                data["cli_valid"].strftime("%b %-d"),
                data["data"].get("temperature_maximum", "M"),
                data["data"].get("temperature_minimum", "M"),
                trace_r(data["data"].get("precip_today", "M")),
                trace_r(data["data"].get("snow_today", "M")),
            )
            tweet = ("%s %s Climate: Hi: %s Lo: %s Precip: %s Snow: %s %s") % (
                data["cli_station"],
                data["cli_valid"].strftime("%b %-d"),
                data["data"].get("temperature_maximum", "M"),
                data["data"].get("temperature_minimum", "M"),
                trace_r(data["data"].get("precip_today", "M")),
                trace_r(data["data"].get("snow_today", "M")),
                url,
            )
            res.append(
                [
                    mess.replace(str(TRACE_VALUE), "Trace"),
                    htmlmess.replace(str(TRACE_VALUE), "Trace"),
                    {
                        "channels": self.get_channels(),
                        "product_id": self.get_product_id(),
                        "twitter": tweet,
                    },
                ]
            )
        return res

    def parse_data(self, section):
        """ Actually do the parsing of this silly format """
        data = {}
        pos = section.find("TEMPERATURE")
        if pos == -1:
            raise CLIException("Failed to find TEMPERATURE, aborting")

        # Strip extraneous spaces
        meat = "\n".join([s.rstrip() for s in section[pos:].split("\n")])
        # replace any 2+ \n with just two
        meat = re.sub(r"\n{2,}", "\n\n", meat)
        sections = meat.split("\n\n")
        for _section in sections:
            lines = _section.split("\n")
            if lines[0] in ["TEMPERATURE (F)", "TEMPERATURE"]:
                parse_temperature(self.regime, lines, data)
            elif lines[0] in ["PRECIPITATION (IN)", "PRECIPITATION"]:
                parse_precipitation(self.regime, lines, data)
            elif lines[0] in ["SNOWFALL (IN)", "SNOWFALL"]:
                parse_snowfall(self.regime, lines, data)
            elif lines[0] in ["SKY COVER"]:
                parse_sky_coverage(lines, data)
            elif lines[0] in ["WIND (MPH)"] and len(lines) > 1:
                parse_wind(lines, data)

        return data

    def parse_cli_headline(self, section):
        """ Figure out when this product is valid for """
        tokens = HEADLINE_RE.findall(section.replace("\n", " "))
        if len(tokens[0][1].split()[0]) == 3:
            myfmt = "%b %d %Y"
        else:
            myfmt = "%B %d %Y"
        cli_valid = datetime.datetime.strptime(tokens[0][1], myfmt).date()
        cli_station = (tokens[0][0]).strip()
        return (cli_valid, cli_station)

    def _sql_data(self, cursor, data):
        """Do an individual data entry."""
        # See what we currently have stored.
        cursor.execute(
            "SELECT product from cli_data where station = %s and valid = %s",
            (data["db_station"], data["cli_valid"]),
        )
        if cursor.rowcount == 1:
            row = cursor.fetchone()
            if self.get_product_id() < row["product"]:
                return
            cursor.execute(
                "DELETE from cli_data WHERE station = %s and valid = %s",
                (data["db_station"], data["cli_valid"]),
            )
        cursor.execute(
            """INSERT into cli_data(
        station, product, valid, high, high_normal, high_record,
        high_record_years, low, low_normal, low_record, low_record_years,
        precip, precip_month, precip_jan1, precip_jul1, precip_normal,
        precip_record,
        precip_record_years, precip_month_normal, snow, snow_month,
        snow_jun1, snow_jul1,
        snow_dec1, precip_dec1, precip_dec1_normal, precip_jan1_normal,
        high_time, low_time, snow_record_years, snow_record,
        snow_jun1_normal, snow_jul1_normal, snow_dec1_normal,
        snow_month_normal, precip_jun1, precip_jun1_normal,
        average_sky_cover,
        resultant_wind_speed, resultant_wind_direction,
        highest_wind_speed, highest_wind_direction,
        highest_gust_speed, highest_gust_direction,
        average_wind_speed)
        VALUES (
        %s, %s, %s, %s, %s, %s,
        %s, %s, %s, %s, %s,
        %s, %s, %s, %s, %s,
        %s,
        %s, %s, %s, %s,
        %s, %s,
        %s, %s, %s, %s,
        %s, %s, %s, %s,
        %s, %s, %s, %s, %s, %s,
        %s,
        %s, %s,
        %s, %s,
        %s, %s,
        %s
        )
        """,
            (
                data["db_station"],
                self.get_product_id(),
                data["cli_valid"],
                data["data"].get("temperature_maximum"),
                data["data"].get("temperature_maximum_normal"),
                data["data"].get("temperature_maximum_record"),
                data["data"].get("temperature_maximum_record_years", []),
                data["data"].get("temperature_minimum"),
                data["data"].get("temperature_minimum_normal"),
                data["data"].get("temperature_minimum_record"),
                data["data"].get("temperature_minimum_record_years", []),
                data["data"].get("precip_today"),
                data["data"].get("precip_month"),
                data["data"].get("precip_jan1"),
                data["data"].get("precip_jul1"),
                data["data"].get("precip_today_normal"),
                data["data"].get("precip_today_record"),
                data["data"].get("precip_today_record_years", []),
                data["data"].get("precip_month_normal"),
                data["data"].get("snow_today"),
                data["data"].get("snow_month"),
                data["data"].get("snow_jun1"),
                data["data"].get("snow_jul1"),
                data["data"].get("snow_dec1"),
                data["data"].get("precip_dec1"),
                data["data"].get("precip_dec1_normal"),
                data["data"].get("precip_jan1_normal"),
                data["data"].get("temperature_maximum_time"),
                data["data"].get("temperature_minimum_time"),
                data["data"].get("snow_today_record_years", []),
                data["data"].get("snow_today_record"),
                data["data"].get("snow_jun1_normal"),
                data["data"].get("snow_jul1_normal"),
                data["data"].get("snow_dec1_normal"),
                data["data"].get("snow_month_normal"),
                data["data"].get("precip_jun1"),
                data["data"].get("precip_jun1_normal"),
                data["data"].get("average_sky_cover"),
                data["data"].get("resultant_wind_speed"),
                data["data"].get("resultant_wind_direction"),
                data["data"].get("highest_wind_speed"),
                data["data"].get("highest_wind_direction"),
                data["data"].get("highest_gust_speed"),
                data["data"].get("highest_gust_direction"),
                data["data"].get("average_wind_speed"),
            ),
        )

    def sql(self, cursor):
        """Do the database update!"""
        for entry in self.data:
            self._sql_data(cursor, entry)
            update_iemaccess(cursor, entry)


def parser(text, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """ Provide back CLI objects based on the parsing of this text """
    # Careful here, see if we have two CLIs in one product!
    return CLIProduct(text, utcnow, ugc_provider, nwsli_provider)



"""Test pyiem.tracker."""
# pylint: disable=redefined-outer-name
import datetime
from datetime import timezone

import pytest
from pyiem.tracker import TrackerEngine, loadqc
from pyiem.network import Table as NetworkTable
from pyiem.util import get_dbconn


@pytest.fixture
def pcursor():
    """Database cursor."""
    dbconn = get_dbconn("portfolio")
    return dbconn.cursor()


@pytest.fixture
def icursor():
    """Database cursor."""
    dbconn = get_dbconn("iem")
    return dbconn.cursor()


def test_loadqc(pcursor):
    """Make sure we exercise the loadqc stuff"""
    q = loadqc()
    assert not q
    q = loadqc(cursor=pcursor)
    assert not q
    pcursor.execute(
        """
        INSERT into tt_base(s_mid, sensor, status, entered) VALUES
        ('BOGUS', 'tmpf', 'OPEN', '2019-03-27')
    """
    )
    q = loadqc(cursor=pcursor, date=datetime.date(2019, 3, 27))
    assert q


def test_workflow(pcursor, icursor):
    """ Test that we can do stuff! """
    sid1 = "XXX"
    sid2 = "YYY"
    pnetwork = "xxxxxx"
    nt = NetworkTable(None)
    nt.sts[sid1] = dict(
        name="XXX Site Name", network="IA_XXXX", tzname="America/Chicago"
    )
    nt.sts[sid2] = dict(
        name="YYY Site Name", network="IA_XXXX", tzname="America/Chicago"
    )
    valid = datetime.datetime.utcnow()
    valid = valid.replace(tzinfo=timezone.utc)
    threshold = valid - datetime.timedelta(hours=3)
    obs = {
        sid1: {"valid": valid},
        sid2: {"valid": valid - datetime.timedelta(hours=6)},
    }
    # Create dummy iem_site_contacts
    pcursor.execute(
        """
        INSERT into iem_site_contacts
        (portfolio, s_mid, email) VALUES (%s, %s, %s)
    """,
        (pnetwork, sid1, "akrherz@localhost"),
    )
    pcursor.execute(
        """
        INSERT into iem_site_contacts
        (portfolio, s_mid, email) VALUES (%s, %s, %s)
    """,
        (pnetwork, sid2, "root@localhost"),
    )
    # Create some dummy tickets
    pcursor.execute(
        """
        INSERT into tt_base (portfolio, s_mid, subject,
        status, author) VALUES (%s, %s, %s, %s, %s) RETURNING id
    """,
        (pnetwork, sid1, "FIXME PLEASE OPEN", "OPEN", "mesonet"),
    )
    pcursor.execute(
        """
        INSERT into tt_base (portfolio, s_mid, subject,
        status, author) VALUES (%s, %s, %s, %s, %s) RETURNING id
    """,
        (pnetwork, sid1, "FIXME PLEASE CLOSED", "CLOSED", "mesonet"),
    )
    tracker = TrackerEngine(icursor, pcursor)
    tracker.process_network(obs, pnetwork, nt, threshold)
    tracker.send_emails(really_send=False)
    assert len(tracker.emails) == 1

    tracker.emails = {}
    obs[sid1]["valid"] = valid - datetime.timedelta(hours=6)
    obs[sid2]["valid"] = valid
    tracker.process_network(obs, pnetwork, nt, threshold)
    tracker.send_emails(really_send=False)
    assert len(tracker.emails) == 2

    tracker.emails = {}
    obs[sid1]["valid"] = valid - datetime.timedelta(hours=6)
    obs[sid2]["valid"] = valid
    tracker.process_network(obs, pnetwork, nt, threshold)
    tracker.send_emails(really_send=False)
    assert not tracker.emails



"""Calendar Plot"""
import os
from collections import OrderedDict
import datetime
import calendar

import numpy as np
import matplotlib.colors as mpcolors
from matplotlib.patches import Rectangle
import matplotlib.image as mpimage
from pyiem.plot.use_agg import plt
from pyiem.plot.colormaps import get_cmap
from pyiem.plot.util import fitbox, fontscale

DATADIR = os.sep.join([os.path.dirname(__file__), "..", "data"])


def iemlogo(fig):
    """Place the IEM Logo"""
    fn = "%s/%s" % (DATADIR, "logo.png")
    if not os.path.isfile(fn):
        return
    logo = mpimage.imread(fn)
    y0 = fig.get_figheight() * 100.0 - logo.shape[0] - 5
    fig.figimage(logo, 5, y0, zorder=3)


def _compute_bounds(sts, ets):
    """figure out our monthly calendar bounding boxes"""
    now = sts
    months = []
    lastmonth = -1
    while now <= ets:
        if now.month != lastmonth:
            months.append(now)
            lastmonth = now.month
        now += datetime.timedelta(days=1)

    bounds = OrderedDict()
    # 1x1
    vpadding = 0.015
    hpadding = 0.01
    if len(months) == 1:
        cols = 1
        rows = 1
    # 1x2
    elif len(months) < 3:
        cols = 2
        rows = 1
    # 2x2
    elif len(months) <= 4:
        cols = 2
        rows = 2
    # 3x3
    elif len(months) <= 9:
        cols = 3
        rows = 3
    # 3x4
    else:
        cols = 3
        rows = 4

    monthtotalwidth = 1.0 / float(cols)
    monthtotalheight = 0.86 / float(rows)
    monthwidth = monthtotalwidth - 2 * hpadding
    monthheight = monthtotalheight - 2 * vpadding

    gx = 0
    gy = 0.9  # upper left corners here
    for i, month in enumerate(months):
        col = i % cols
        row = int(i / cols)
        llx = gx + col * monthtotalwidth
        lly = gy - (row + 1) * monthtotalheight
        bounds[month] = [
            llx + hpadding,
            lly + vpadding,
            monthwidth,
            monthheight,
        ]
    return bounds


def _do_cell(axes, now, data, row, dx, dy, kwargs):
    """Do what work is necessary within the cell"""
    val = data.get(now, dict()).get("val")
    cellcolor = (
        "None"
        if kwargs.get("norm") is None or val is None
        else kwargs["cmap"](kwargs["norm"]([val]))[0]
    )
    offx = (now.weekday() + 1) if now.weekday() != 6 else 0
    cellcolor = data.get(now, dict()).get("cellcolor", cellcolor)
    rect = Rectangle(
        (offx * dx, 0.9 - (row + 1) * dy),
        dx,
        dy,
        zorder=(2 if val is None else 3),
        facecolor=cellcolor,
        edgecolor="tan" if val is None else "k",
    )
    axes.add_patch(rect)
    if val is None:
        return
    color = "k"
    if not isinstance(cellcolor, str):  # this is a string comp here
        color = (
            "k"
            if (
                cellcolor[0] * 256 * 0.299
                + cellcolor[1] * 256 * 0.587
                + cellcolor[2] * 256 * 0.114
            )
            > 186
            else "white"
        )
    color = data[now].get("color", color)
    # We need to translate the axes NDC coordinates back to the figure coords
    bbox = axes.get_position()
    sdx = (bbox.x1 - bbox.x0) * dx
    sdy = (bbox.y1 - bbox.y0) * dy
    x0 = bbox.x0 + offx * sdx
    ytop = bbox.y0 + (bbox.y1 - bbox.y0) * 0.9
    y0 = ytop - (row + 1) * sdy
    fitbox(
        plt.gcf(),
        val,
        x0,
        x0 + sdx,
        y0,
        y0 + sdy * 0.55,
        ha="center",
        va="center",
        color=color,
    )


def _do_month(month, axes, data, in_sts, in_ets, kwargs):
    """Place data on this axes"""
    axes.get_xaxis().set_visible(False)
    axes.get_yaxis().set_visible(False)
    pos = axes.get_position()
    ndcheight = pos.y1 - pos.y0
    ndcwidth = pos.x1 - pos.x0

    fitbox(
        plt.gcf(),
        month.strftime("%B %Y"),
        pos.x0,
        pos.x1,
        pos.y1,
        pos.y1 + 0.028,
        ha="center",
    )

    axes.add_patch(
        Rectangle(
            (0.0, 0.90), 1, 0.1, zorder=2, facecolor="tan", edgecolor="tan"
        )
    )

    sts = datetime.date(month.year, month.month, 1)
    ets = (sts + datetime.timedelta(days=35)).replace(day=1)

    calendar.setfirstweekday(calendar.SUNDAY)
    weeks = len(calendar.monthcalendar(month.year, month.month))
    now = sts
    row = 0
    dy = 0.9 / float(weeks)
    dx = 1.0 / 7.0
    for i, dow in enumerate(["SUN", "MON", "TUE", "WED", "THU", "FRI", "SAT"]):
        axes.text(
            1.0 / 7.0 * (i + 0.5),
            0.94,
            dow,
            fontsize=fontscale(ndcwidth / 8.0 * 0.4),
            ha="center",
            va="center",
        )
    while now < ets:
        # Is this Sunday?
        if now.weekday() == 6 and now != sts:
            row += 1
        if now < in_sts or now > in_ets:
            now += datetime.timedelta(days=1)
            continue
        offx = (now.weekday() + 1) if now.weekday() != 6 else 0
        axes.text(
            offx * dx + 0.01,
            0.9 - row * dy - 0.01,
            str(now.day),
            fontsize=fontscale(ndcheight / 5.0 * 0.25),
            color="tan",
            va="top",
        )
        _do_cell(axes, now, data, row, dx, dy, kwargs)
        now += datetime.timedelta(days=1)


def calendar_plot(sts, ets, data, **kwargs):
    """Create a plot that looks like a calendar

    Args:
      sts (datetime.date): start date of this plot
      ets (datetime.date): end date of this plot (inclusive)
      data (dict[dict]): dictionary with keys of dates and dicts for
        `val` value and optionally `color` for color
      kwargs (dict):
        heatmap (bool): background color for cells based on `val`, False
        cmap (str): color map to use for norm
    """
    bounds = _compute_bounds(sts, ets)
    # Compute the number of month calendars we need.

    # We want 'square' boxes for each month's calendar, 4x3
    fig = plt.figure(figsize=(10.24, 7.68))
    if "fontsize" not in kwargs:
        kwargs["fontsize"] = 12
        if len(bounds) < 3:
            kwargs["fontsize"] = 18
        elif len(bounds) < 5:
            kwargs["fontsize"] = 16
        elif len(bounds) < 10:
            kwargs["fontsize"] = 14
    if kwargs.get("heatmap", False):
        kwargs["cmap"] = get_cmap(kwargs.get("cmap", "viridis"))
        maxval = -1000
        for key in data:
            if data[key]["val"] > maxval:
                maxval = data[key]["val"]
        # Need at least 3 slots
        maxval = 5 if maxval < 5 else maxval
        # Need to have more colors than bins
        kwargs["norm"] = mpcolors.BoundaryNorm(
            np.arange(0, maxval, int(maxval / 255.0) + 1), kwargs["cmap"].N
        )
    for month in bounds:
        ax = fig.add_axes(bounds[month])
        _do_month(month, ax, data, sts, ets, kwargs)

    iemlogo(fig)
    title = kwargs.get("title")
    if title is not None:
        fitbox(fig, title, 0.1, 0.99, 0.95, 0.99)

    subtitle = kwargs.get("subtitle")
    if subtitle is not None:
        if subtitle.find("\n") > 0:  # Allow more room
            fitbox(fig, subtitle, 0.1, 0.99, 0.909, 0.949)
        else:
            fitbox(fig, subtitle, 0.1, 0.99, 0.925, 0.945)

    return fig



"""Is our reference hackery usable."""

from pyiem import reference


def test_reference():
    """Can we import everything from our API."""
    for name in reference._onthefly_dict:
        res = getattr(reference, name, None)
        # is a dictionary
        assert isinstance(res, dict)
        # has keys
        assert res.keys()



"""test IEMRE stuff"""
import datetime

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

import numpy as np
from pyiem.util import utc, get_dbconn
from pyiem import iemre


def test_ncname():
    """Test the responses for get_names."""
    assert iemre.get_daily_ncname(2020) is not None
    assert iemre.get_hourly_ncname(2020) is not None
    assert iemre.get_daily_mrms_ncname(2020) is not None


def test_get_table():
    """Test get_table."""
    d1 = datetime.date(2000, 8, 1)
    assert iemre.get_table(d1) == "iemre_daily_2000"
    d2 = utc(2000, 9, 1, 12)
    assert iemre.get_table(d2) == "iemre_hourly_200009"


def test_get_gid():
    """Can we get a gid?"""
    assert iemre.get_gid(-96, 44) is not None


def test_forecast_grids():
    """Test getting and setting grids from the future."""
    pgconn = get_dbconn("iemre")
    cursor = pgconn.cursor()
    valid = datetime.date(2029, 12, 1)
    cursor.execute(
        """
        DELETE from iemre_daily_forecast WHERE valid = %s
    """,
        (valid,),
    )
    cursor.execute(
        """
        INSERT into iemre_daily_forecast
        (gid, valid, high_tmpk, low_tmpk, p01d, rsds)
        select gid, %s, random(), random(),
        random(), random() from iemre_grid LIMIT 100
    """,
        (valid,),
    )
    ds = iemre.get_grids(valid, cursor=cursor, table="iemre_daily_forecast")
    assert "high_tmpk" in ds
    assert "bogus" not in ds

    iemre.set_grids(valid, ds, cursor=cursor, table="iemre_daily_forecast")


def test_get_grids():
    """Can we get grids?"""
    pgconn = get_dbconn("iemre")
    cursor = pgconn.cursor()
    valid = utc(2019, 12, 1, 1)
    cursor.execute(
        """
        DELETE from iemre_hourly_201912 WHERE valid = %s
    """,
        (valid,),
    )
    cursor.execute(
        """
        DELETE from iemre_hourly_201912 WHERE valid = %s
    """,
        (valid + datetime.timedelta(days=1),),
    )
    cursor.execute(
        """
        INSERT into iemre_hourly_201912
        (gid, valid, tmpk, dwpk, uwnd, vwnd, p01m)
        select gid, %s, random(), null, random(),
        random(), random() from iemre_grid LIMIT 100
    """,
        (valid,),
    )
    ds = iemre.get_grids(valid, varnames="tmpk", cursor=cursor)
    assert "tmpk" in ds
    assert "bogus" not in ds
    ds = iemre.get_grids(valid, cursor=cursor)
    assert np.isnan(ds["dwpk"].values.max())

    iemre.set_grids(valid, ds, cursor=cursor)
    iemre.set_grids(valid + datetime.timedelta(days=1), ds, cursor=cursor)


def test_simple():
    """ Get nulls for right and top values """
    i, j = iemre.find_ij(iemre.EAST, iemre.NORTH)
    assert i is None
    assert j is None

    i, j = iemre.find_ij(iemre.WEST, iemre.SOUTH)
    assert i == 0
    assert j == 0


def test_hourly_offset():
    """ Compute the offsets """
    ts = utc(2013, 1, 1, 0, 0)
    offset = iemre.hourly_offset(ts)
    assert offset == 0

    ts = utc(2013, 1, 1, 6, 0)
    ts = ts.astimezone(ZoneInfo("America/Chicago"))
    offset = iemre.hourly_offset(ts)
    assert offset == 6

    ts = utc(2013, 1, 5, 12, 0)
    offset = iemre.hourly_offset(ts)
    assert offset == 4 * 24 + 12


def test_daily_offset():
    """ Compute the offsets """
    ts = utc(2013, 1, 1, 0, 0)
    offset = iemre.daily_offset(ts)
    assert offset == 0

    ts = datetime.date(2013, 2, 1)
    offset = iemre.daily_offset(ts)
    assert offset == 31

    ts = utc(2013, 1, 5, 12, 0)
    offset = iemre.daily_offset(ts)
    assert offset == 4



"""SIGMET"""
from collections import defaultdict

from pyiem.nws.products.sigmet import parser, compute_esol
from pyiem.util import utc, get_test_file


def mydict():
    """return dict."""
    return dict(lon=-85.50, lat=42.79)


NWSLI_PROVIDER = defaultdict(mydict)


def test_190503_badgeom():
    """This SIGMET produced a traceback in prod."""
    utcnow = utc(2019, 5, 3, 18, 25)
    tp = parser(
        get_test_file("SIGMETS/SIGC_badgeom.txt"),
        utcnow,
        nwsli_provider=NWSLI_PROVIDER,
    )
    assert len(tp.sigmets) == 4


def test_170815_pywwa_issue3():
    """This example was in pyWWA issues list, so lets test here"""
    utcnow = utc(2015, 9, 30, 16, 56)

    tp = parser(
        get_test_file("SIGMETS/SIGE.txt"),
        utcnow,
        nwsli_provider=NWSLI_PROVIDER,
    )
    assert len(tp.sigmets) == 4


def test_150930_sigak2():
    """Got an error with this product"""
    utcnow = utc(2015, 9, 30, 16, 56)
    tp = parser(get_test_file("SIGMETS/SIGAK2.txt"), utcnow)
    assert not tp.sigmets


def test_150921_sigpas():
    """Got an error with this product"""
    utcnow = utc(2015, 9, 21, 10, 57)
    tp = parser(get_test_file("SIGMETS/SIGPAS.txt"), utcnow)
    assert len(tp.sigmets) == 1


def test_150917_cancel():
    """Don't error out on a CANCELs SIGMET"""
    utcnow = utc(2015, 9, 17, 0, 0)
    tp = parser(get_test_file("SIGMETS/SIGPAP_cancel.txt"), utcnow)
    assert not tp.sigmets


def test_compute_esol():
    """ Test our algo on either side of a line """
    pts = [[0, 0], [5, 0]]
    pts = compute_esol(pts, 111)
    print(pts)
    assert abs(pts[0][0] - 0.00) < 0.01
    assert abs(pts[0][1] - 1.00) < 0.01
    assert abs(pts[1][0] - 5.00) < 0.01
    assert abs(pts[1][1] - 1.00) < 0.01
    assert abs(pts[2][0] - 5.00) < 0.01
    assert abs(pts[2][1] - -1.00) < 0.01
    assert abs(pts[3][0] - 0.00) < 0.01
    assert abs(pts[3][1] - -1.00) < 0.01
    assert abs(pts[4][0] - 0.00) < 0.01
    assert abs(pts[4][1] - 1.00) < 0.01


def test_150915_line():
    """ See about parsing a SIGMET LINE """
    utcnow = utc(2015, 9, 15, 2, 55)
    ugc_provider = {}
    nwsli_provider = {
        "MSP": dict(lon=-83.39, lat=44.45),
        "MCW": dict(lon=-85.50, lat=42.79),
    }
    tp = parser(
        get_test_file("SIGMETS/SIGC_line.txt"),
        utcnow,
        ugc_provider,
        nwsli_provider,
    )
    assert abs(tp.sigmets[0].geom.area - 0.47) < 0.01


def test_150915_isol():
    """ See about parsing a SIGMET ISOL """
    utcnow = utc(2015, 9, 12, 23, 55)
    ugc_provider = {}
    nwsli_provider = {
        "FTI": dict(lon=-83.39, lat=44.45),
        "CME": dict(lon=-85.50, lat=42.79),
    }
    tp = parser(
        get_test_file("SIGMETS/SIGC_ISOL.txt"),
        utcnow,
        ugc_provider,
        nwsli_provider,
    )
    assert abs(tp.sigmets[0].geom.area - 0.30) < 0.01
    assert abs(tp.sigmets[1].geom.area - 0.30) < 0.01


def test_150915_nospace():
    """ See about parsing a SIGMET that has no spaces """
    utcnow = utc(2015, 9, 15, 15, 41)
    tp = parser(get_test_file("SIGMETS/SIGAX.txt"), utcnow)
    assert abs(tp.sigmets[0].geom.area - 23.47) < 0.01


def test_140907_circle():
    """ See about parsing a SIGMET that is circle? """
    utcnow = utc(2014, 9, 6, 22, 15)
    tp = parser(get_test_file("SIGMETS/SIGP0H.txt"), utcnow)
    assert abs(tp.sigmets[0].geom.area - 11.70) < 0.01


def test_140813_line():
    """ See about parsing a SIGMET that is a either side of line """
    utcnow = utc(2014, 8, 12, 13, 15)
    tp = parser(get_test_file("SIGMETS/SIGP0A_line.txt"), utcnow)
    assert abs(tp.sigmets[0].geom.area - 4.32) < 0.01


def test_140815_cancel():
    """ See about parsing a SIGMET that is a either side of line """
    utcnow = utc(2014, 8, 15, 23, 41)
    tp = parser(get_test_file("SIGMETS/SIG_cancel.txt"), utcnow)
    assert not tp.sigmets


def test_sigaoa():
    """ SIGAOA """
    utcnow = utc(2014, 8, 11, 19, 15)
    tp = parser(get_test_file("SIGMETS/SIGA0A.txt"), utcnow)
    assert abs(tp.sigmets[0].geom.area - 24.35) < 0.01


def test_sigaob():
    """ See about parsing 50E properly """
    utcnow = utc(2014, 8, 11, 19, 15)
    tp = parser(get_test_file("SIGMETS/SIGA0B.txt"), utcnow)
    assert not tp.sigmets


def test_50e():
    """ See about parsing 50E properly """
    utcnow = utc(2014, 8, 11, 18, 55)
    ugc_provider = {}
    nwsli_provider = {
        "ASP": dict(lon=-83.39, lat=44.45),
        "ECK": dict(lon=-82.72, lat=43.26),
        "GRR": dict(lon=-85.50, lat=42.79),
    }

    tp = parser(
        get_test_file("SIGMETS/SIGE3.txt"),
        utcnow,
        ugc_provider,
        nwsli_provider,
    )
    assert abs(tp.sigmets[0].geom.area - 2.15) < 0.01


def test_sigc():
    """ See about parsing SIGC """
    utcnow = utc(2014, 8, 11, 16, 55)
    ugc_provider = {}
    nwsli_provider = {}
    for sid in (
        "MSL,SJI,MLU,LIT,BTR,LEV,LCH,IAH,YQT,SAW,SAT,DYC,AXC,"
        "ODI,DEN,TBE,ADM,JCT,INK,ELP"
    ).split(","):
        nwsli_provider[sid] = dict(lon=-99, lat=45)

    tp = parser(
        get_test_file("SIGMETS/SIGC.txt"), utcnow, ugc_provider, nwsli_provider
    )
    j = tp.get_jabbers("http://localhost", "http://localhost")
    assert tp.sigmets[0].ets == utc(2014, 8, 11, 18, 55)
    ans = "KKCI issues SIGMET 62C for AL MS LA AR till 1855 UTC"
    assert j[0][0] == ans
    ans = (
        "KKCI issues SIGMET 63C for LA TX AND MS LA TX CSTL WTRS till 1855 UTC"
    )
    assert j[1][0] == ans


def test_sigpat():
    """ Make sure we don't have another failure with geom parsing """
    utcnow = utc(2014, 8, 11, 12, 34)
    tp = parser(get_test_file("SIGMETS/SIGPAT.txt"), utcnow)
    j = tp.get_jabbers("http://localhost", "http://localhost")
    assert abs(tp.sigmets[0].geom.area - 33.71) < 0.01
    assert tp.sigmets[0].sts == utc(2014, 8, 11, 12, 35)
    assert tp.sigmets[0].ets == utc(2014, 8, 11, 16, 35)
    assert j[0][0] == "PHFO issues SIGMET TANGO 1 till 1635 UTC"



"""Test plots made by pyiem.plot.geoplot"""
import datetime
import tempfile
import os
import copy

import pytest
import matplotlib.colors as mpcolors
import numpy as np
from pyiem import reference
from pyiem import plot
from pyiem.plot.geoplot import windrose
from pyiem.plot import (
    MapPlot,
    centered_bins,
    pretty_bins,
    load_bounds,
    load_pickle_pd,
    load_pickle_geo,
)


def test_depreciated():
    """Test that our windrose is depreciated."""
    with pytest.deprecated_call():
        windrose("")


def test_invalid_file():
    """Test that we don't error out on an invalid filename."""
    assert load_bounds("this shall not work") is None
    assert load_pickle_pd("this shall not work") is None
    assert not load_pickle_geo("this shall not work")


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_issue217():
    """See what happens with our logo on very scaled figure."""
    mp = MapPlot(nocaption=True, figsize=(6.00, 3.35))
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_issue98_labelbar():
    """Sometimes our label bar sucks."""
    mp = MapPlot(
        title="Proportional Colorbar with some rotation",
        sector="iowa",
        nocaption=True,
    )
    cmap = copy.copy(plot.maue())
    cmap.set_under("white")
    cmap.set_over("black")
    clevs = np.arange(0, 1.0, 0.1)
    clevs[-1] = 3.987654
    norm = mpcolors.BoundaryNorm(clevs, cmap.N)
    mp.plot_values(
        [-94, -92, -91, -92],
        [42, 41, 43, 42.4],
        ["0.5", "0.25", "1.0", "5.0"],
        color=cmap(norm([0.5, 0.25, 1.0, 5.0])),
        showmarker=True,
    )
    mp.draw_colorbar(clevs, cmap, norm, spacing="proportional")
    return mp.fig


def test_savefile():
    """Can we properly save a file."""
    mp = MapPlot()
    tmpfd = tempfile.NamedTemporaryFile(delete=False)
    mp.postprocess(filename=tmpfd.name)
    assert os.path.isfile(tmpfd.name)


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_illinois():
    """Produce a plot that doesn't suck"""
    mp = MapPlot(sector="state", state="IL", nocaption=True)
    mp.draw_cwas()
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_issue292_nws_fill_ugcs():
    """Test that fill_ugcs works for nws sector view."""
    mp = MapPlot(sector="nws", title="Four Counties", nocaption=True)
    data = {"IAC001": 10, "AKC013": 20, "HIC001": 30, "PRC001": 40}
    mp.fill_ugcs(data)
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_fill_ugcs_color():
    """Provide an explicit color to fill_ugcs"""
    mp = MapPlot(
        sector="cwa", cwa="DMX", title="Three Counties", nocaption=True
    )
    data = {"IAC001": 10, "IAC003": 20, "IAC135": 30}
    color = {"IAC001": "#FF0000", "IAC003": "black"}
    mp.fill_ugcs(data, color=color, nocbar=True)
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_dep():
    """Produce a plot with the DEP logo on it."""
    mp = MapPlot(sector="state", state="IA", nocaption=True, logo="dep")
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.7)
def test_usdm():
    """Can we plot the current USDM"""
    mp = MapPlot(sector="conus", nocaption=True)
    mp.draw_usdm(valid=datetime.date(2018, 5, 7), hatched=True, filled=False)
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_hexbin():
    """See if we can do hexbin OKish"""
    mp = MapPlot(
        sector="north_america", continentalcolor="white", nocaption=True
    )
    lons = np.arange(-100, -80, 0.25)
    lats = np.arange(40, 50, 0.25)
    vals = np.linspace(0, 1, lats.shape[0] * lons.shape[0]).reshape(
        [lats.shape[0], lons.shape[0]]
    )
    lons, lats = np.meshgrid(lons, lats)
    mp.hexbin(
        lons.flatten(),
        lats.flatten(),
        vals.flatten(),
        np.arange(0, 1, 0.1),
        cmap="jet",
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.2)
def test_pcolormesh():
    """See if we can do pcolormesh OKish"""
    mp = MapPlot(
        sector="custom",
        north=43,
        east=-80,
        west=-96,
        south=38,
        projection=reference.EPSG[2163],
        continentalcolor="white",
        nocaption=True,
    )
    lons = np.arange(-100, -80, 0.25)
    lats = np.arange(40, 50, 0.25)
    vals = np.linspace(0, 1, lats.shape[0] * lons.shape[0]).reshape(
        [lats.shape[0], lons.shape[0]]
    )
    lons, lats = np.meshgrid(lons, lats)
    mp.pcolormesh(lons, lats, vals, np.arange(0, 1, 0.1))
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.3)
def test_conus():
    """See if we can plot albers"""
    mp = MapPlot(
        sector="custom",
        title="EPSG: 5070 Albers",
        north=reference.CONUS_NORTH + 1,
        east=reference.CONUS_EAST - 12,
        west=reference.CONUS_WEST + 14,
        south=reference.CONUS_SOUTH,
        projection=reference.EPSG[5070],
        continentalcolor="white",
        nocaption=True,
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.3)
def test_conus2():
    """Map the conus in LEA"""
    mp = MapPlot(
        sector="custom",
        title="EPSG: 2163 Lambert Azimuthal Equal Area",
        north=reference.CONUS_NORTH + 1,
        east=reference.CONUS_EAST - 12,
        west=reference.CONUS_WEST + 14,
        south=reference.CONUS_SOUTH,
        projection=reference.EPSG[2163],
        continentalcolor="white",
        nocaption=True,
    )
    return mp.fig


def test_pretty_bins():
    """Test that we get nice pretty bins!"""
    a = pretty_bins(-1, 10)
    assert abs(a[-1] - 10.5) < 0.01


def test_centered_bins():
    """See that we can compute some nice centered bins"""
    a = centered_bins(10)
    assert a[0] == -10
    a = centered_bins(55)
    assert a[0] == -56
    a = centered_bins(99)
    assert a[0] == -100
    a = centered_bins(99, bins=9)
    assert a[0] == -99
    a = centered_bins(100, on=100)
    assert a[0] == 0
    a = centered_bins(0.9)
    assert abs(a[-1] - 1.2) < 0.001
    a = centered_bins(1.2888)
    assert abs(a[-1] - 1.6) < 0.001


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_michigan():
    """See what we do with Michigan"""
    mp = MapPlot(sector="state", state="MI", nocaption=True)
    mp.contourf(
        np.arange(-84, -75),
        np.arange(36, 45),
        np.arange(9),
        np.arange(9),
        clevlabels=["a", "b", "c", "d", "e", "f", "g", "h", "i"],
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_drawcities():
    """Draw Cities"""
    mp = MapPlot(
        title="Fill and Draw Cities",
        subtitle="This is my subtitle",
        continentalcolor="blue",
        sector="iowa",
        nocaption=True,
    )
    mp.drawcities()
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_drawrandomtext():
    """See if we can handle the fun that is drawing random text"""
    mp = MapPlot(
        sector="iowa",
        title="Fun Text, here and there",
        continentalcolor="white",
        debug=True,
        nocaption=True,
    )
    mp.plot_values(
        [-94, -92, -91, -92],
        [42, 41, 43, 42.4],
        ["One", "Two\nTwo", "Three\nThree\nThree", "Four\nFour\nFour\nFour"],
        showmarker=True,
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.2)
def test_drawiowawfo():
    """Iowa Contour Plot"""
    mp = MapPlot(sector="iowawfo", title="Iowa Contour plot", nocaption=True)
    mp.contourf(
        np.arange(-94, -85),
        np.arange(36, 45),
        np.arange(9),
        np.arange(9),
        clevlabels=["a", "b", "c", "d", "e", "f", "g", "h", "i"],
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.3)
def test_fillstates():
    """Can we fill states"""
    data = {"AK": 10, "HI": 30, "IA": 40, "NY": 80}
    mp = MapPlot(
        sector="nws",
        title="Fill AK, HI, IA, NY States",
        subtitle="test_fillstates",
        nocaption=True,
    )
    mp.fill_states(data, ilabel=True)
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.25)
def test_drawcounties():
    """draw counties on the map"""
    mp = MapPlot(sector="midwest", title="Counties", nocaption=True)
    mp.drawcounties()
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.25)
def test_drawcounties_cornbelt():
    """draw counties on the map"""
    mp = MapPlot(sector="cornbelt", title="Counties", nocaption=True)
    mp.drawcounties()
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.25)
def test_drawcounties_iailin():
    """draw IA IL IN masked"""
    mp = MapPlot(sector="iailin", title="Counties", nocaption=True)
    mp.contourf(
        np.arange(-94, -85),
        np.arange(36, 45),
        np.arange(9),
        np.arange(9),
        clevlabels=["a", "b", "c", "d", "e", "f", "g", "h", "i"],
    )
    mp.drawcounties()
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=1.0)
def test_climdiv():
    """Run tests agains the fill_climdiv"""
    mp = MapPlot(sector="conus", title="Climate Divisions", nocaption=True)
    data = {"IAC001": 10, "MNC001": 20, "NMC001": 30}
    mp.fill_climdiv(data, ilabel=True)
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_colorbar():
    """Run tests against the colorbar algorithm"""
    mp = MapPlot(sector="iowa", nocaption=True)
    cmap = copy.copy(plot.maue())
    cmap.set_under("white")
    clevs = list(range(0, 101, 10))
    norm = mpcolors.BoundaryNorm(clevs, cmap.N)
    mp.drawcities()
    mp.draw_colorbar(clevs, cmap, norm)
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_colorbar2():
    """draw a colorbar"""
    mp = MapPlot(sector="iowa", nocaption=True)
    cmap = plot.maue()
    clevs = list(range(0, 101, 10))
    clevlabels = [
        "One",
        "Three",
        "Blahh",
        "Longest",
        "Five",
        "Six",
        "Ten",
        "Fourty",
        100000,
        "Hi\nHo",
        100,
    ]
    norm = mpcolors.BoundaryNorm(clevs, cmap.N)
    mp.draw_colorbar(
        clevs, cmap, norm, clevlabels=clevlabels, extend="neither"
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_colorbar3():
    """draw another colorbar"""
    mp = MapPlot(sector="iowa", nocaption=True)
    cmap = copy.copy(plot.maue())
    cmap.set_over("black")
    clevs = [0, 100, 250, 500, 1000, 2000, 20000]
    norm = mpcolors.BoundaryNorm(clevs, cmap.N)
    mp.draw_colorbar(
        clevs, cmap, norm, title="Erosion $kg/m^2$", spacing="uniform"
    )
    return mp.fig


# as of writing, python2.7 failure tolerance of 1.45
@pytest.mark.mpl_image_compare(tolerance=1.6)
def test_drawugcs():
    """test drawing of UGCS"""
    mp = MapPlot(
        sector="conus", title="Counties, 3 filled in Iowa", nocaption=True
    )
    mp.fill_ugcs({"IAC001": 10, "IAC003": 20, "IAC005": 30})
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=1.0)
def test_drawugcs2():
    """3 filled zones"""
    mp = MapPlot(
        sector="iowa",
        title="Zones, 3 filled in Iowa, label",
        subtitle="test_drawugcs2",
        nocaption=True,
    )
    mp.fill_ugcs({"IAZ001": 10, "IAZ003": 20, "IAZ005": 30}, ilabel=True)
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_ugcs_lwx():
    """Ensure that we can plot some counties in LWX CWA."""
    mp = MapPlot(
        sector="cwa",
        cwa="LWX",
        title="Two Maryland Counties",
        subtitle="test_ugcs_lwx",
        nocaption=True,
    )
    labels = {"MDC003": "MDC003", "MDC033": "MDC033"}
    mp.fill_ugcs(
        {"MDC003": 1, "MDC033": 40},
        bins=list(range(0, 101, 10)),
        labels=labels,
        ilabel=True,
        extend="min",
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_ugcs_lwx_zones():
    """Ensure that we can plot some zones in LWX CWA."""
    mp = MapPlot(
        sector="cwa",
        cwa="LWX",
        title="Two Maryland Counties",
        subtitle="test_ugcs_lwx",
        nocaption=True,
    )
    labels = {"MDZ001": "MDZ001", "MDZ008": "MDZ008"}
    mp.fill_ugcs(
        {"MDZ001": 1, "MDZ008": 40},
        bins=list(range(0, 101, 10)),
        labels=labels,
        ilabel=True,
        extend="min",
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_ugcs_withcustomlabels():
    """Fill ugcs with provided labels."""
    mp = MapPlot(
        sector="iowa",
        title="Zones, 3 filled in Iowa, custom label",
        subtitle="test_drawugcs2",
        nocaption=True,
    )
    labels = {"IAZ001": "IAZ001", "IAZ003": "IAZ003"}
    bins = list(range(24))
    clevlabels = [""] * 24
    clevlabels[0] = "mid"
    clevlabels[12] = "noon"
    mp.fill_ugcs(
        {"IAZ001": 1, "IAZ003": 4, "IAZ005": 12},
        bins=bins,
        labels=labels,
        ilabel=True,
        clevstride=12,
        clevlabels=clevlabels,
        extend="neither",
    )
    return mp.fig


def test_filter_functions():
    """Make sure our filter functions are doing what we want!"""
    mp = MapPlot(sector="iowa")
    assert plot.state_filter(mp, "IAC001", dict())
    assert not plot.state_filter(mp, "MNC001", dict())
    mp = MapPlot(cwa="DMX")
    assert plot.state_filter(mp, "IAC001", dict())


@pytest.mark.mpl_image_compare(tolerance=0.15)
def test_states():
    """Exercise the state plotting routines"""
    mp = MapPlot(sector="state", state="CA", nocaption=True)
    assert mp.state == "CA"
    return mp.fig


# high tolerance due to python2.7 issue I don't wish to deal with now.
@pytest.mark.mpl_image_compare(tolerance=4.0)
def test_cwa_with_custom_masking():
    """Exercise the cwa plotting routines"""
    mp = MapPlot(sector="cwa", cwa="DLH", nocaption=True)
    mp.contourf(
        np.arange(-94, -89),
        np.arange(45, 50),
        np.arange(5),
        np.arange(5),
        clevlabels=["a", "b", "c", "d", "e"],
        clip_on=False,
    )
    mp.draw_cwas()
    mp.draw_mask(sector="conus")
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_cwa():
    """Exercise the cwa plotting routines"""
    mp = MapPlot(sector="cwa", cwa="MKX", nocaption=True)
    mp.contourf(
        np.arange(-94, -89),
        np.arange(40, 45),
        np.arange(5),
        np.arange(5),
        clevlabels=["a", "b", "c", "d", "e"],
    )
    mp.draw_cwas()
    assert mp.cwa == "MKX"
    return mp.fig


def test_colorramps():
    """make sure our colorramps are happy"""
    c = plot.james()
    assert c.N == 12
    c = plot.james2()
    assert c.N == 12
    c = plot.whitebluegreenyellowred()
    assert c.N == 236
    c = plot.nwssnow()
    assert c.N == 11


@pytest.mark.mpl_image_compare(tolerance=0.2)
def test_overlap():
    """ Do some checking of our overlaps logic """
    mp = MapPlot(sector="midwest", continentalcolor="white", nocaption=True)
    lons = np.linspace(-99, -90, 100)
    lats = np.linspace(38, 44, 100)
    vals = lats
    labels = ["%.2f" % (s,) for s in lats]
    mp.plot_values(lons, lats, vals, fmt="%.2f", labels=labels)
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_barbs():
    """Testing the plotting of wind barbs"""
    mp = MapPlot(continentalcolor="white", nocaption=True)
    data = [
        dict(lat=41.5, lon=-96, tmpf=50, dwpf=30, sknt=10, drct=100),
        dict(lat=42.0, lon=-95.5, tmpf=50, dwpf=30, sknt=20, drct=200),
    ]
    mp.plot_station(data, fontsize=12)
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_stationplot():
    """Testing the plotting of wind barbs"""
    mp = MapPlot(continentalcolor="white", nocaption=True)
    data = [
        dict(lat=41.5, lon=-96, tmpf=50, dwpf=30, id="BOOI4"),
        dict(lat=42.0, lon=-95.5, tmpf=50, dwpf=30, id="CSAI4"),
    ]
    mp.plot_station(data, fontsize=12)
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.2)
def test_scatter():
    """ Test scatter plots """
    mp = MapPlot(
        sector="midwest",
        title="Should see 100 dots",
        subtitle="test_scatter",
        nocaption=True,
    )
    mp.scatter(
        np.linspace(-99, -94, 100),
        np.linspace(40, 45, 100),
        np.arange(100),
        np.arange(0, 101, 10),
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=4)
def test_contourf():
    """ Test the contourf plot with labels specified """
    mp = MapPlot(sector="iowa", nocaption=True)
    mp.contourf(
        np.arange(-94, -89),
        np.arange(40, 45),
        np.arange(5),
        np.arange(5),
        clevlabels=["a", "b", "c", "d", "e"],
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_textplot():
    """ Can we plot text and place labels on them """
    mp = MapPlot(sector="iowa", nocaption=True)
    mp.plot_values(np.arange(-99, -94), np.arange(40, 45), np.arange(5))
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_textplot2():
    """plot values on a map"""
    mp = MapPlot(sector="iowa", nocaption=True)
    mp.plot_values(
        np.arange(-99, -94),
        np.arange(40, 45),
        np.arange(5),
        labels=range(5, 11),
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=1.0)
def test_plot():
    """ Exercise the API """
    mp = MapPlot(sector="midwest", nocaption=True)
    # test the translation of JSJ to SJU
    mp.fill_cwas({"DMX": 80, "MKX": 5, "JSJ": 30, "AJK": 40}, units="no units")
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=1.0)
def test_plot2():
    """ Exercise NWS plot API """
    mp = MapPlot(sector="nws", continentalcolor="white", nocaption=True)
    mp.fill_cwas(
        {"DMX": 80, "MKX": 5, "SJU": 30, "AJK": 40, "HFO": 50},
        units="NWS Something or Another",
        ilabel=True,
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.63)
def test_plot22():
    """plot cwas that are filled"""
    mp = MapPlot(sector="iowa", continentalcolor="white", nocaption=True)
    mp.fill_cwas(
        {"DMX": 80, "MKX": 5, "SJU": 30, "AJK": 40, "HFO": 50},
        units="NWS Something or Another",
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.2)
def test_plot3():
    """ Exercise climdiv plot API """
    mp = MapPlot(sector="iowa", nocaption=True)
    mp.fill_climdiv(
        {"IAC001": 80, "AKC003": 5, "HIC003": 30, "AJK": 40, "HFO": 50}
    )
    return mp.fig


@pytest.mark.mpl_image_compare(tolerance=0.1)
def test_alaska():
    """See that Alaska plots nicely."""
    mp = MapPlot(sector="state", state="AK", nocaption=True)
    return mp.fig



"""
Utility Functions that are common to our scripts, I hope
"""
import json
import os
import sys
import re

import gdata.gauth
import gdata.sites.client as sclient
from oauth2client.service_account import ServiceAccountCredentials
from httplib2 import Http
from googleapiclient.discovery import build
import smartsheet
from pyiem.util import LOG

CONFIG_FN = "/opt/datateam/config/mytokens.json"
NUMBER_RE = re.compile(r"^[-+]?\d*\.\d+$|^\d+$")
CLEANVALUE_COMPLAINED = []
CLEANVALUE_XREF = {"NA": "n/a", "dnc": "did not collect"}


def save_config(config, filename=None):
    """ Save the configuration to disk """
    if filename is None:
        filename = CONFIG_FN
    json.dump(
        config,
        open(filename, "w"),
        sort_keys=True,
        indent=4,
        separators=(",", ": "),
    )


def get_config(filename=None):
    """ Load a JSON Configuration"""
    if filename is None:
        filename = CONFIG_FN
    if not os.path.isfile(filename):
        sys.stderr.write(
            f"cscap_utils.get_config({filename}) File Not Found.\n"
        )
        return None
    return json.load(open(filename))


def cleanvalue(val):
    """cleanup the mess that is found in the Google Sheets for values

    Args:
      val (str): The value to clean up

    Returns:
      the cleaned value!
    """
    if val is None or val.strip() == "":
        return None
    if NUMBER_RE.match(val):
        return float(val)
    if CLEANVALUE_XREF.get(val):
        return CLEANVALUE_XREF[val]
    if val.lower() in [
        "did not collect",
        ".",
        "n/a",
        "clay",
        "silty clay",
        "silty clay loam",
        "clay loam",
        "sandy clay loam",
        "silt loam",
        "silty loam",
        "sandy loam",
        "sandy clay",
        "sand",
        "loam",
        "silt",
        "loamy sand",
    ]:
        return val.lower()
    if val.find("%") > -1:
        val = val.replace("%", "")
        if NUMBER_RE.match(val):
            return float(val)
    if val.find("<") > -1:
        return "< %s" % (val.replace("<", "").strip(),)
    if val not in CLEANVALUE_COMPLAINED:
        LOG.info(
            "cscap_utils.cleanvalue(%s) is unaccounted for, return None",
            repr(val),
        )
        CLEANVALUE_COMPLAINED.append(val)
    return None


def translate_years(val):
    """ Convert X ('YY-'YY) into an array"""
    if val.find("-") > 0:
        tokens = re.findall("[0-9]+", val)
        one = int(tokens[0])
        two = int(tokens[1])
        one = (1900 + one) if one > 50 else (2000 + one)
        two = (1900 + two) if two > 50 else (2000 + two)
        return range(one, two + 1)
    tokens = re.findall("[0-9]+", val)
    return [int("%s%s" % ("19" if int(t) > 50 else "20", t)) for t in tokens]


def get_xref_plotids(drive):
    """Dictionary of Sites to PlotID keys

    Args:
      drive: authorized Google Drive API client

    Returns:
      dict
    """
    res = drive.files().list(q="title contains 'Plot Identifiers'").execute()
    data = {}
    for item in res["items"]:
        if item["mimeType"] != "application/vnd.google-apps.spreadsheet":
            continue
        siteid = item["title"].split()[0]
        data[siteid] = item["id"]
    return data


def get_sites_client(config, site="sustainablecorn"):
    """ Return an authorized sites client """

    token = gdata.gauth.OAuth2Token(
        client_id=config["appauth"]["client_id"],
        client_secret=config["appauth"]["app_secret"],
        user_agent="daryl.testing",
        scope=config["googleauth"]["scopes"],
        refresh_token=config["googleauth"]["refresh_token"],
    )

    sites_client = sclient.SitesClient(site=site)
    token.authorize(sites_client)
    return sites_client


def get_driveclient(config, project="cscap"):
    """ Return an authorized apiclient """
    return get_googleapiclient(config, project, "drive", "v2")


def get_sheetsclient(config, project="cscap"):
    """ Return an authorized apiclient """
    return get_googleapiclient(config, project, "sheets", "v4")


def get_googleapiclient(config, project, ns, v):
    """Helper to get an authorized googleapiclient

    Args:
      config (dict): provider of configuration
      project (str): the project used within config
      ns (str): google endpoint to use
      v (str): google endpoint version to use
    """
    cred = ServiceAccountCredentials.from_json_keyfile_dict(
        config[project]["service_account"],
        scopes=["https://www.googleapis.com/auth/drive"],
    )
    http_auth = cred.authorize(Http())
    return build(ns, v, http=http_auth)


def get_folders(drive):
    """Return a dict of Google Drive Folders"""
    f = {}

    # Whoa, just because maxResults=999 and the returned items is less
    # than 999, it does not mean the list was complete
    folders = (
        drive.files()
        .list(
            q="mimeType = 'application/vnd.google-apps.folder'", maxResults=999
        )
        .execute()
    )
    folder_list = folders["items"]
    i = 0
    while "nextPageToken" in folders:
        folders = (
            drive.files()
            .list(
                pageToken=folders["nextPageToken"],
                q="mimeType = 'application/vnd.google-apps.folder'",
                maxResults=999,
            )
            .execute()
        )
        folder_list = folder_list + folders["items"]
        i += 1
        if i > 10:
            LOG.info("get_folders iterator reached 10, aborting")
            break

    for _, item in enumerate(folder_list):
        f[item["id"]] = dict(title=item["title"], parents=[], basefolder=None)
        for parent in item["parents"]:
            f[item["id"]]["parents"].append(parent["id"])

    for thisfolder in f:
        # title = f[thisfolder]['title']
        if not f[thisfolder]["parents"]:
            continue
        parentfolder = f[thisfolder]["parents"][0]
        if parentfolder not in f:
            LOG.info("ERROR: parentfolder: %s not in f", parentfolder)
            continue
        while parentfolder in f and len(f[parentfolder]["parents"]) > 0:
            parentfolder = f[parentfolder]["parents"][0]
        f[thisfolder]["basefolder"] = parentfolder
    return f


def get_ssclient(config):
    """Return a smartsheet enabled client"""
    return smartsheet.Smartsheet(config["ss_access_token"])



"""
 We do meteorological things, when necessary
"""
import math

import numpy as np
import metpy.calc as mcalc
from metpy.units import masked_array
import pyiem.datatypes as dt


InvalidArguments = Exception


def clearsky_shortwave_irradiance_year(lat, elevation):
    """Compute the Clear Sky Shortwave Irradiance for year in MJ m**-2

    Args:
      lat (float): latitude
      elevation (float): location elevation in meters

    Returns:
      irradiance (list)
    """
    # Mean pressure in kPa
    pa = 101.3 * math.exp((0 - elevation) / 8200.0)
    # TOA radiation Wm**2
    spo = 1360.0
    # assume clearsky
    tau = 0.75
    # julian days
    j = np.arange(1, 366, 1)
    # solar declination
    # delta = asind(0.39785.*sind((278.97+0.9856.*J+1.9165.
    # *sind((356.6+0.9856.*J)))))
    _a = np.sin(np.radians(356.6 + 0.9856 * j))
    _b = np.sin(np.radians(278.97 + 0.9856 * j + 1.9165 * _a))
    delta = np.degrees(np.arcsin(0.39785 * _b))
    data = []
    for jday in j:
        running = 0
        for t in np.arange(0, 12.001, 5.0 / 60.0):
            # acosd((sind(gamma(l))*sind(delta(j))+cosd(gamma(l))*
            # cosd(delta(j))*cosd(15*(t-12))))
            _a = math.cos(np.radians(15 * (t - 12)))
            _b = math.sin(np.radians(lat))
            _c = math.sin(np.radians(delta[jday - 1]))
            theta = np.degrees(
                math.acos(
                    _b * _c
                    + math.cos(np.radians(lat))
                    * math.cos(np.radians(delta[jday - 1]))
                    * _a
                )
            )
            if theta >= 90:
                continue
            m = pa / (101.3 * math.cos(np.radians(theta)))
            direct = spo * tau ** m * math.cos(np.radians(theta))
            diffuse = 0.3 * (1 - tau ** m) * spo * math.cos(np.radians(theta))
            running += (5.0 * 60) * (direct + diffuse)
        data.append((running * 2.0) / 1000000.0)
    return data


def drct(u, v):
    """
    Compute the wind direction given a u and v wind speed

    Args:
      u (dt.speed): u component wind speed
      v (dt.speed): v component wind speed

    Returns:
      dt.direction value
    """
    umps = u.value("MPS")
    vmps = v.value("MPS")
    val = (np.arctan2(umps, vmps) * 180.0 / np.pi) + 180
    return dt.direction(val, "DEG")


def uv(speed, direction):
    """
    Compute the u and v components of the wind
    @param wind speed in whatever units
    @param dir wind direction with zero as north
    @return u and v components
    """
    if not isinstance(speed, dt.speed) or not isinstance(
        direction, dt.direction
    ):
        raise InvalidArguments(
            ("uv() needs speed and direction " "objects as args")
        )
    # Get radian units
    rad = direction.value("RAD")
    if rad is None or speed.value() is None:
        return None, None
    u = (0 - speed.value()) * np.sin(rad)
    v = (0 - speed.value()) * np.cos(rad)
    return (dt.speed(u, speed.get_units()), dt.speed(v, speed.get_units()))


def mcalc_feelslike(tmpf, dwpf, smps):
    """Compute a feels like temperature

    Args:
      temperature (temperature): The dry bulb temperature
      dewpoint (temperature): The dew point temperature
      speed (speed): the wind speed

    Returns:
      temperature (temperature): The feels like temperature
    """
    is_not_scalar = isinstance(tmpf.m, (list, tuple, np.ndarray))
    rh = mcalc.relative_humidity_from_dewpoint(tmpf, dwpf)
    # NB: update this once metpy 0.11 is released
    app = mcalc.apparent_temperature(tmpf, rh, smps)
    if hasattr(app, "mask"):
        if is_not_scalar:
            app[app.mask] = tmpf[app.mask]
        else:
            app = tmpf
    elif hasattr(tmpf, "mask"):
        app = masked_array(app.m, app.units)
        app.mask = tmpf.mask

    return app


def windchill(temperature, speed):
    """Compute the wind chill temperature

    http://www.ofcm.gov/publications/reports2.htm

    Args:
      temperature (temperature): The Air Temperature
      speed (speed): The Wind Speed

    Returns:
      temperature (temperature): The Wind Chill Temperature
    """
    tmpf = temperature.value("F")
    sknt = speed.value("KT")
    wci = (
        35.74
        + 0.6215 * tmpf
        - 35.75 * np.power(sknt, 0.16)
        + 0.4275 * tmpf * np.power(sknt, 0.16)
    )
    wci = np.where(
        np.logical_or(np.less(sknt, 3), np.greater(tmpf, 50)), tmpf, wci
    )
    return dt.temperature(wci, "F")


def heatindex(temperature, polyarg):
    """
    Compute the heat index based on

    Stull, Richard (2000). Meteorology for Scientists and Engineers,
    Second Edition. Brooks/Cole. p. 60. ISBN 9780534372149.

    Another opinion on appropriate equation:
    http://www.wpc.ncep.noaa.gov/html/heatindex_equation.shtml

    http://www.weather.gov/media/ffc/ta_htindx.PDF
    """
    if not isinstance(temperature, dt.temperature):
        raise InvalidArguments("heatindex() needs temperature obj as arg")
    if isinstance(polyarg, dt.temperature):  # We have dewpoint
        polyarg = relh(temperature, polyarg)
    rh = polyarg.value("%")
    t = temperature.value("F")
    t2 = t ** 2
    t3 = t ** 3
    rh2 = rh ** 2
    rh3 = rh ** 3
    hdx = (
        16.923
        + ((1.85212e-1) * t)
        + (5.37941 * rh)
        - ((1.00254e-1) * t * rh)
        + ((9.41695e-3) * t2)
        + ((7.28898e-3) * rh2)
        + ((3.45372e-4) * t2 * rh)
        - ((8.14971e-4) * t * rh2)
        + ((1.02102e-5) * t2 * rh2)
        - ((3.8646e-5) * t3)
        + ((2.91583e-5) * rh3)
        + ((1.42721e-6) * t3 * rh)
        + ((1.97483e-7) * t * rh3)
        - ((2.18429e-8) * t3 * rh2)
        + ((8.43296e-10) * t2 * rh3)
        - ((4.81975e-11) * t3 * rh3)
    )
    hdx = np.where(np.logical_or(np.less(t, 80), np.greater(t, 120)), t, hdx)
    return dt.temperature(hdx, "F")


def dewpoint_from_pq(pressure, mixingratio):
    """
    Compute the Dew Point given a Pressure and Mixing Ratio
    """
    p = pressure.value("hPa")
    mr = mixingratio.value("kg/kg")
    e = (p * mr) / (0.622 + mr)
    b = 26.66082 - np.log(e)
    t = (b - (b * b - 223.1986) ** 0.5) / 0.0182758048
    return dt.temperature(t, "K")


def dewpoint(temperature, relhumid):
    """
    Compute Dew Point given a temperature and RH%
    """
    tmpk = temperature.value("K")
    _relh = relhumid.value("%")
    dwpk = tmpk / (1 + 0.000425 * tmpk * -(np.log10(_relh / 100.0)))
    return dt.temperature(dwpk, "K")


def relh(temperature, _dewpoint):
    """
    Compute Relative Humidity based on a temperature and dew point
    """
    # Get temperature in Celsius
    tmpc = temperature.value("C")
    dwpc = _dewpoint.value("C")

    e = 6.112 * np.exp((17.67 * dwpc) / (dwpc + 243.5))
    es = 6.112 * np.exp((17.67 * tmpc) / (tmpc + 243.5))
    _relh = (e / es) * 100.00
    return dt.humidity(_relh, "%")


def mixing_ratio(_dewpoint):
    """Compute the mixing ratio

    Args:
      dewpoint (temperature): Dew Point temperature

    Returns:
      mixing ratio
    """
    dwpc = _dewpoint.value("C")
    e = 6.112 * np.exp((17.67 * dwpc) / (dwpc + 243.5))
    return dt.mixingratio(0.62197 * e / (1000.0 - e), "KG/KG")


def gdd(high, low, base=50.0, ceiling=86.0):
    """Compute Growing Degree Days

    Args:
      high (temperature): High Temperature
      low (temperature): Low Temperature
      base (int): Base to use in GDD Computation (F)
      ceiling (int): Ceiling to use in GDD Computation (F)

    Returns:
      float value for GDDs
    """
    highf = high.value("F")
    lowf = low.value("F")
    highf = np.ma.where(np.ma.less(highf, base), base, highf)
    lowf = np.ma.where(np.ma.less(lowf, base), base, lowf)
    highf = np.ma.where(np.ma.greater(highf, ceiling), ceiling, highf)
    lowf = np.ma.where(np.ma.greater(lowf, ceiling), ceiling, lowf)
    res = (highf + lowf) / 2.0 - 50.0
    if res.shape == [1]:
        return res[0]
    return res



"""Support library for the IEM Reanalysis code

.. data:: SOUTH

    Latitude of the southern edge of the IEM Reanalysis.


"""
import string
import random
from datetime import timezone, datetime

from affine import Affine
import numpy as np
import xarray as xr
from six import string_types
from pyiem.util import get_dbconn

# 1/8 degree grid, grid cell is the lower left corner
SOUTH = 23.0
WEST = -126.0
NORTH = 50.0
EAST = -65.0

DX = 0.125
DY = 0.125
# hard coding these to prevent flakey behaviour with dynamic computation
NX = 488  # int((EAST - WEST) / DX)
NY = 216  # int((NORTH - SOUTH) / DY)
XAXIS = np.arange(WEST, EAST, DX)
YAXIS = np.arange(SOUTH, NORTH, DY)
AFFINE = Affine(DX, 0.0, WEST, 0.0, 0 - DY, NORTH)
MRMS_AFFINE = Affine(0.01, 0.0, WEST, 0.0, -0.01, NORTH)


def get_table(valid):
    """Figure out which table should be used for given valid.

    Args:
      valid (datetime or date):  which time is of interest

    Returns:
      str tablename
    """
    # careful here, a datetime is not an instance of date
    if isinstance(valid, datetime):
        table = "iemre_hourly_%s" % (
            valid.astimezone(timezone.utc).strftime("%Y%m"),
        )
    else:
        table = f"iemre_daily_{valid.year}"
    return table


def set_grids(valid, ds, cursor=None, table=None):
    """Update the database with a given ``xarray.Dataset``.

    Args:
      valid (datetime or date): If datetime, save hourly, if date, save daily
      ds (xarray.Dataset): The xarray dataset to save
      cursor (database cursor, optional): cursor to use for queries
      table (str,optional): hard coded database table to use to set the data
        on.  Usually dynamically computed.
    """
    table = table if table is not None else get_table(valid)
    commit = cursor is None
    if cursor is None:
        pgconn = get_dbconn("iemre")
        cursor = pgconn.cursor()
    # see that we have database entries, otherwise create them
    cursor.execute(
        f"SELECT valid from {table} WHERE valid = %s LIMIT 1", (valid,)
    )
    insertmode = True
    if cursor.rowcount == 1:
        # Update mode, we do some massive tricks :/
        temptable = "".join(random.choices(string.ascii_uppercase, k=24))
        insertmode = False
        update_cols = ", ".join(
            ["%s = $%i" % (v, i + 1) for i, v in enumerate(ds)]
        )
        arg = "$%i" % (len(ds) + 1,)
        # approximate table size is 10 MB
        cursor.execute("SET temp_buffers = '100MB'")
        cursor.execute(
            f"CREATE UNLOGGED TABLE {temptable} AS SELECT * from {table} "
            f"WHERE valid = '{valid}'"
        )
        cursor.execute(f"CREATE INDEX on {temptable}(gid)")
        cursor.execute(
            (
                f"PREPARE pyiem_iemre_plan as UPDATE {temptable} "
                f"SET {update_cols} WHERE gid = {arg}"
            )
        )
    else:
        # Insert mode
        insert_cols = ", ".join(["%s" % (v,) for v in ds])
        percents = ", ".join(["$%i" % (i + 2,) for i in range(len(ds))])
        cursor.execute(
            f"PREPARE pyiem_iemre_plan as INSERT into {table} "
            f"(gid, valid, {insert_cols}) VALUES($1, '{valid}', {percents})"
        )
    sql = "execute pyiem_iemre_plan (%s)" % (",".join(["%s"] * (len(ds) + 1)),)

    def _n(val):
        """Prevent nan"""
        return None if np.isnan(val) else float(val)

    # Implementation notes: xarray iteration was ~25 secs, loading into memory
    # instead is a few seconds :/
    pig = {v: ds[v].values for v in ds}
    for y in range(ds.dims["y"]):
        for x in range(ds.dims["x"]):
            arr = [_n(pig[v][y, x]) for v in ds]
            if insertmode:
                arr.insert(0, y * NX + x)
            else:
                arr.append(y * NX + x)
            cursor.execute(sql, arr)
    if not insertmode:
        # Undo our hackery above
        cursor.execute(f"DELETE from {table} WHERE valid = '{valid}'")
        cursor.execute(f"INSERT into {table} SELECT * from {temptable}")
        cursor.execute(f"DROP TABLE {temptable}")
    # If we generate a cursor, we should save it
    if commit:
        cursor.close()
        pgconn.commit()
    else:
        cursor.execute("""DEALLOCATE pyiem_iemre_plan""")


def get_grids(valid, varnames=None, cursor=None, table=None):
    """Fetch grid(s) from the database, returning xarray.

    Args:
      valid (datetime or date): If datetime, load hourly, if date, load daily
      varnames (str or list,optional): Which variables to fetch from database,
        defaults to all available
      cursor (database cursor,optional): cursor to use for query
      table (str,optional): Hard coded table to fetch data from, useful in the
        case of forecast data.

    Returns:
      ``xarray.Dataset``"""
    table = table if table is not None else get_table(valid)
    if cursor is None:
        pgconn = get_dbconn("iemre")
        cursor = pgconn.cursor()
    # rectify varnames
    if isinstance(varnames, string_types):
        varnames = [varnames]
    # Compute variable names
    cursor.execute(
        "SELECT column_name FROM information_schema.columns "
        "WHERE table_schema = 'public' AND table_name = %s and "
        "column_name not in ('gid', 'valid')",
        (table,),
    )
    use_columns = []
    for row in cursor:
        if not varnames or row[0] in varnames:
            use_columns.append(row[0])
    colsql = ",".join(use_columns)
    cursor.execute(
        f"SELECT (gid / %s)::int as y, gid %% %s as x, {colsql} "
        f"from {table} WHERE valid = %s",
        (NX, NX, valid),
    )
    data = dict((key, np.full((NY, NX), np.nan)) for key in use_columns)
    for row in cursor:
        for i, col in enumerate(use_columns):
            data[col][row[0], row[1]] = row[2 + i]
    ds = xr.Dataset(
        dict((key, (["y", "x"], data[key])) for key in data),
        coords={"lon": (["x"], XAXIS), "lat": (["y"], YAXIS)},
    )
    return ds


def get_dailyc_ncname():
    """Return the filename of the daily climatology netcdf file"""
    return "/mesonet/data/iemre/iemre_dailyc.nc"


def get_daily_ncname(year):
    """Get the daily netcdf filename for the given year"""
    return f"/mesonet/data/iemre/{year}_iemre_daily.nc"


def get_dailyc_mrms_ncname():
    """Get the MRMS daily climatology filename"""
    return "/mesonet/data/iemre/iemre_mrms_dailyc.nc"


def get_daily_mrms_ncname(year):
    """Get the daily netcdf MRMS filename for the given year"""
    return f"/mesonet/data/iemre/{year}_iemre_mrms_daily.nc"


def get_hourly_ncname(year):
    """Get the daily netcdf filename for the given year"""
    return f"/mesonet/data/iemre/{year}_iemre_hourly.nc"


def daily_offset(ts):
    """ Compute the timestamp index in the netcdf file """
    # In case ts is passed here as a datetime.date object
    ts = datetime(ts.year, ts.month, ts.day)
    base = ts.replace(
        month=1, day=1, hour=0, minute=0, second=0, microsecond=0
    )
    days = (ts - base).days
    return int(days)


def hourly_offset(dtobj):
    """Return time index for given timestamp

    Args:
      dtobj (datetime): datetime, if no tzinfo, we assume it is UTC

    Returns:
      int time index in the netcdf file
    """
    if dtobj.tzinfo and dtobj.tzinfo != timezone.utc:
        dtobj = dtobj.astimezone(timezone.utc)
    base = dtobj.replace(
        month=1, day=1, hour=0, minute=0, second=0, microsecond=0
    )
    seconds = (dtobj - base).total_seconds()
    return int(seconds / 3600.0)


def find_ij(lon, lat):
    """Compute which grid cell this lon, lat resides within"""
    if lon < WEST or lon >= EAST or lat < SOUTH or lat >= NORTH:
        return None, None

    i = np.digitize([lon], XAXIS)[0] - 1
    j = np.digitize([lat], YAXIS)[0] - 1

    return i, j


def get_gid(lon, lat):
    """Compute the grid id for the given location."""
    i, j = find_ij(lon, lat)
    if i is None:
        return None
    return j * NX + i



"""Test Local Storm Report parsing."""
# pylint: disable=redefined-outer-name

import pytest
from pyiem.util import get_test_file, get_dbconn
from pyiem.nws.products.lsr import parser, parse_lsr


@pytest.fixture
def dbcursor():
    """Database cursor."""
    return get_dbconn("postgis").cursor()


def test_issue331_state_channels():
    """Test that we assign new state based channels to LSRs."""
    prod = parser(get_test_file("LSR/LSRFSD.txt"))
    j = prod.get_jabbers("")
    channels = j[0][2]["channels"].split(",")
    assert "LSR.SD.TORNADO" in channels
    assert "LSR.SD" in channels


def test_unknown_units():
    """Test what happens when we have unknown units."""
    prod = parser(get_test_file("LSR/LSRJAN_fakeunits.txt"))
    assert str(prod.lsrs[0])
    j = prod.get_jabbers("")
    ans = (
        "At 11:37 AM CST, Yokena [Warren Co, MS] AMATEUR RADIO reports "
        "HAIL of 40 C. FAKED LSR FOR TESTING PURPOSES. "
        "#JAN/201912161737/201912161737"
    )
    assert j[0][2]["twitter"] == ans


def test_summary():
    """Test that our summary logic works."""
    prod = parser(get_test_file("LSR/LSR.txt"))
    j = prod.get_jabbers("")
    prod.lsrs[0].duplicate = True
    prod.duplicates = 1
    j2 = prod.get_jabbers("")
    assert (len(j) - len(j2)) == 1


def test_issue320_badmnd():
    """Test we can deal with a bad MND timestamp header."""
    prod = parser(get_test_file("LSR/LSRTBW_badmnd.txt"))
    assert prod.z is not None


def test_issue320_reallybadmnd():
    """Test what happens when we have no workable MND."""
    prod = parser(get_test_file("LSR/LSRTBW_badmnd.txt").replace("NOV 12", ""))
    assert prod.z is None
    assert not prod.lsrs


def test_200913_nounits():
    """Test that we properly handle LSRs without units."""
    prod = parser(get_test_file("LSR/LSRCRP.txt"))
    assert not prod.warnings


def test_empty():
    """Test that we can handle an empty LSR."""
    prod = parser(get_test_file("LSR/LSR_empty.txt"))
    res = prod.get_temporal_domain()
    assert res[0] is None and res[1] is None


def test_too_short():
    """Test a psuedo too short LSR."""
    prod = parser(get_test_file("LSR/LSRDVN_old.txt"))
    res = parse_lsr(prod, "")
    assert res is None


def test_duplicate():
    """Test that we can mark a LSR as a duplicate."""
    prod = parser(get_test_file("LSR/LSRDVN_old.txt"))
    j = prod.get_jabbers("")
    assert len(j) == 1
    prod.lsrs[0].duplicate = True
    j = prod.get_jabbers("")
    assert not j


def test_issue277_oldlsr():
    """Test the tweet generated from a delayed LSR."""
    prod = parser(get_test_file("LSR/LSRDVN_old.txt"))
    assert len(prod.lsrs) == 1
    j = prod.get_jabbers("")
    ans = (
        "[Delayed Report] On Aug 10, at 12:28 PM CDT, 3 WSW Cedar Rapids "
        "[Linn Co, IA] PUBLIC reports "
        "TSTM WND GST of E130 MPH. ROOF REMOVED FROM SMALL "
        "STRIP MALL. SOME EXTERIOR WALLS COLLAPSED. "
        "#DVN/202008101728/202008101728"
    )
    assert j[0][2]["twitter"] == ans


def test_sql(dbcursor):
    """Test that we can insert into the database."""
    prod = parser(get_test_file("LSR/LSRDVN_old.txt"))
    assert prod.lsrs
    for lsr in prod.lsrs:
        lsr.sql(dbcursor)
        assert dbcursor.rowcount == 1


def test_issue170_nan():
    """How are we handling LSRs that have bad NAN magnitudes."""
    prod = parser(get_test_file("LSR/LSRJAN_NAN.txt"))
    assert len(prod.warnings) == 1
    assert not prod.lsrs


def test_issue61_future():
    """Can we properly warn on a product from the future."""
    prod = parser(get_test_file("LSR/LSRGSP_future.txt"))
    assert len(prod.warnings) == 1
    assert not prod.lsrs



"""Test NHC."""

from pyiem.nws.products.nhc import parser as nhcparser
from pyiem.util import get_test_file


def test_issue317_nowarning():
    """Test that no warning is emitted for a tropical product from TJSJ."""
    prod = nhcparser(get_test_file("tropical/TCPSP4.txt"))
    prod.get_jabbers("http://localhost", "http://localhost")
    assert not prod.warnings


def test_170618_potential():
    """New TCP type"""
    prod = nhcparser(get_test_file("TCPAT2.txt"))
    assert not prod.warnings
    j = prod.get_jabbers("http://localhost", "http://localhost")
    ans = (
        "National Hurricane Center issues ADVISORY 2 "
        "for POTENTIAL TROPICAL CYCLONE TWO "
        "http://localhost?pid=201706190300-KNHC-WTNT32-TCPAT2"
    )
    assert j[0][0] == ans


def test_160905_correction():
    """See that a product correction does not trip us"""
    prod = nhcparser(get_test_file("TCPAT4.txt"))
    assert not prod.warnings
    j = prod.get_jabbers("http://localhost", "http://localhost")
    ans = (
        "National Hurricane Center issues ADVISORY 28A "
        "for POST-TROPICAL CYCLONE HERMINE INTERMEDIATE "
        "http://localhost?pid=201609041200-KNHC-WTNT34-TCPAT4"
    )
    assert j[0][0] == ans



"""
 Something to store UGC information!
"""
import re
import datetime
from collections import OrderedDict

from pyiem.exceptions import UGCParseException

UGC_RE = re.compile(
    r"^(([A-Z]?[A-Z]?[C,Z]?[0-9]{3}[>\-]\s?\n?)+)([0-9]{6})-$", re.M
)


def ugcs_to_text(ugcs):
    """ Convert a list of UGC objects to a textual string """
    states = OrderedDict()
    geotype = "counties"
    for ugc in ugcs:
        code = str(ugc)
        state_abbr = code[:2]
        if code[2] == "Z":
            geotype = "forecast zones"
        if state_abbr not in states:
            states[state_abbr] = []
        if ugc.name is None:
            name = f"(({code}))"
        else:
            name = ugc.name
        states[state_abbr].append(name)

    txt = []
    for st in states.keys():
        states[st].sort()
        part = " %s [%s]" % (", ".join(states[st]), st)
        if len(part) > 350:
            if st == "LA" and geotype == "counties":
                geotype = "parishes"
            part = " %s %s in [%s]" % (len(states[st]), geotype, st)
        txt.append(part)

    return (" and".join(txt)).strip()


def str2time(text, valid):
    """Convert a string that is the UGC product expiration to a valid
    datetime
    @param text string to convert
    @param valid datetime instance
    """
    if text in ["000000", "123456"]:
        return None
    day = int(text[:2])
    hour = int(text[2:4])
    minute = int(text[4:])
    if day < 5 and valid.day > 25:  # Next month
        valid = valid + datetime.timedelta(days=25)
    # elif day > 25 and valid.day < 5: # previous month
    #    valid = valid - datetime.timedelta(days=25)

    return valid.replace(day=day, hour=hour, minute=minute)


def parse(text, valid, ugc_provider=None):
    """Helper method that parses text and yields UGC and expiration time
    @param text to parse
    @param valid is the issue time of the product this text was found in
    @param ugc_provider of UGC objects
    """
    if ugc_provider is None:
        ugc_provider = dict()

    def _construct(code):
        return ugc_provider.get(code, UGC(code[:2], code[2], code[3:]))

    ugcs = []
    expire = None
    tokens = UGC_RE.findall(text)
    if not tokens:
        return ugcs, expire
    # TODO: perhaps we should be more kind when we find products with this
    #       formatting error, but we can recover.  Note that typically the
    #       UGC codes are the same, but the product expiration time may be off
    # if len(tokens) == 2 and tokens[0] == tokens[1]:
    #    pass
    if len(tokens) > 1:
        raise UGCParseException(
            "More than 1 UGC encoding in text:\n%s\n" % (str(tokens),)
        )

    parts = re.split("-", tokens[0][0].replace(" ", "").replace("\n", ""))
    expire = str2time(tokens[0][2], valid)
    state_code = ""
    for i, part in enumerate(parts):
        if i == 0:
            if len(part) >= 6:
                ugc_type = part[2]
            else:
                # This is bad encoding
                raise UGCParseException(
                    ('WHOA, bad UGC encoding detected "%s"')
                    % ("-".join(parts),)
                )
        this_part = parts[i].strip()
        if len(this_part) == 6:  # We have a new state ID
            state_code = this_part[:3]
            ugcs.append(_construct(this_part))
        elif len(this_part) == 3:  # We have an individual Section
            ugcs.append(
                _construct(
                    "%s%s%s" % (state_code[:2], state_code[2], this_part)
                )
            )
        elif len(this_part) > 6:  # We must have a > in there somewhere
            new_parts = re.split(">", this_part)
            first_part = new_parts[0]
            second_part = new_parts[1]
            if len(first_part) > 3:
                state_code = first_part[:3]
            first_val = int(first_part[-3:])
            last_val = int(second_part)
            if ugc_type == "C":
                for j in range(0, last_val + 2 - first_val, 2):
                    str_code = "%03i" % (first_val + j,)
                    ugcs.append(
                        _construct(
                            "%s%s%s"
                            % (state_code[:2], state_code[2], str_code)
                        )
                    )
            else:
                for j in range(first_val, last_val + 1):
                    str_code = "%03i" % (j,)
                    ugcs.append(
                        _construct(
                            "%s%s%s"
                            % (state_code[:2], state_code[2], str_code)
                        )
                    )
    return ugcs, expire


class UGC:
    """Representation of a single UGC"""

    def __init__(self, state, geoclass, number, name=None, wfos=None):
        """
        Constructor for UGC instances
        """
        self.state = state
        self.geoclass = geoclass
        self.number = int(number)
        self.name = name
        self.wfos = wfos if wfos is not None else []

    def __str__(self):
        """ Override str() """
        return "%s%s%03i" % (self.state, self.geoclass, self.number)

    def __repr__(self):
        """ Override repr() """
        return "%s%s%03i" % (self.state, self.geoclass, self.number)

    def __eq__(self, other):
        """ Compare this UGC with another """
        return (
            self.state == other.state
            and self.geoclass == other.geoclass
            and self.number == other.number
        )

    def __ne__(self, other):
        """ Compare this UGC with another """
        return not self == other

    __hash__ = None  # unhashable



"""Pilot Reports (PIREP)

This module attempts to process and store atomic data from PIREPs.  These are
encoded products that look like so:

  UBUS01 KMSC 221700
  EAU UA /OV EAU360030/TM 1715/FL350/TP B737/TB CONT LGT-MOD CHOP =
  EHY UA /OV MBW253036 /TM 1729 /FL105 /TP C206 /SK FEW250 /TA M06
  /TB NEG /RM SMTH=

Unfortunately, there is not much documentation of this format and the feed of
this data contains a bunch of formatting errors.

"""
from enum import Enum
import datetime
import re
import math

from pydantic import BaseModel
import pyiem.nws.product as product
from pyiem.datatypes import distance
from pyiem.util import html_escape, LOG

OV_LATLON = re.compile(
    (
        r"\s?(?P<lat>[0-9]{3,4})(?P<latsign>[NS])"
        r"\s?(?P<lon>[0-9]{3,5})(?P<lonsign>[EW])"
    )
)
OV_LOCDIR = re.compile(
    r".*?(?P<loc>[A-Z0-9]{3,4})\s?(?P<dir>[0-9]{3})(?P<dist>[0-9]{3})"
)
OV_TWOLOC = re.compile(
    r"(?P<loc1>[A-Z0-9]{3,4})\s?-\s?(?P<loc2>[A-Z0-9]{3,4})"
)
OV_OFFSET = re.compile(
    (
        r"(?P<dist>[0-9]{1,3})\s?"
        "(?P<dir>NORTH|EAST|SOUTH|WEST|N|NNE|NE|ENE|E|ESE|"
        r"SE|SSE|S|SSW|SW|WSW|W|WNW|NW|NNW)\s+(OF )?(?P<loc>[A-Z0-9]{3,4})"
    )
)

DRCT2DIR = {
    "N": 0,
    "NNE": 22.5,
    "NE": 45,
    "ENE": 67.5,
    "E": 90,
    "ESE": 112.5,
    "SE": 135,
    "SSE": 157.5,
    "S": 180,
    "SSW": 202.5,
    "SW": 225,
    "WSW": 247.5,
    "W": 270,
    "WNW": 292.5,
    "NW": 305,
    "NNW": 327.5,
    "NORTH": 0,
    "EAST": 90,
    "SOUTH": 180,
    "WEST": 270,
}


class Priority(str, Enum):
    """Types of reports."""

    def __str__(self):
        """When we want the str repr."""
        return str(self.value)

    UA = "UA"
    UUA = "UUA"


class PilotReport(BaseModel):
    """ A Pilot Report. """

    base_loc: str = None
    text: str = None
    priority: Priority = None
    latitude: float = None
    longitude: float = None
    valid: datetime.datetime = None
    cwsu: str = None
    aircraft_type: str = None
    is_duplicate: bool = False


class Pirep(product.TextProduct):
    """ Class for parsing and representing Space Wx Products. """

    def __init__(
        self, text, utcnow=None, ugc_provider=None, nwsli_provider=None
    ):
        """ constructor """
        product.TextProduct.__init__(
            self,
            text,
            utcnow=utcnow,
            ugc_provider=ugc_provider,
            nwsli_provider=nwsli_provider,
        )
        self.reports = []
        self.parse_reports()

    def parse_reports(self):
        """Actually do the parsing of the product that generates the reports
        stored within the self.reports list"""
        txt = (
            self.unixtext
            if self.unixtext[:2] != "\001\n"
            else self.unixtext[2:]
        )

        lines = txt.split("\n")
        # There may be an AWIPSID in line 3 or silly aviation control char
        pos = 3 if len(lines[2]) < 10 or lines[2].startswith("\x1e") else 2
        meat = "".join(lines[pos:])
        for report in meat.split("="):
            if report.strip() == "":
                continue
            res = self.process_pirep(" ".join(report.strip().split()))
            if res is not None:
                self.reports.append(res)

    def process_pirep(self, report):
        """ Convert this report text into an actual PIREP object """
        _pr = PilotReport()
        _pr.text = report

        for i, token in enumerate(report.split("/")):
            token = token.strip()
            # First token is always priority
            if i == 0:
                if len(token) > 10:
                    LOG.info("Aborting as not-PIREP? |%s|", report)
                    return
                if token.find(" UUA") > 0:
                    _pr.priority = Priority.UUA
                else:
                    _pr.priority = Priority.UA
                parts = token.split()
                if len(parts) == 2:
                    _pr.base_loc = parts[0]
                    if len(_pr.base_loc) == 4 and _pr.base_loc[0] == "K":
                        _pr.base_loc = _pr.base_loc[1:]
                continue
            # Aircraft Type
            if token.startswith("TP "):
                _pr.aircraft_type = token[3:]

            # Location
            if token.startswith("OV "):
                dist = 0
                bearing = 0
                therest = token[3:]
                if len(therest) == 3:
                    loc = therest
                elif therest.startswith("FINAL RWY"):
                    loc = report[:8].split()[0]
                    if len(loc) == 4 and loc[0] == "K":
                        loc = loc[1:]
                elif len(therest) == 4:
                    if therest[0] == "K":
                        loc = therest[1:]
                    else:
                        loc = therest
                elif re.match(OV_OFFSET, therest):
                    d = re.match(OV_OFFSET, therest).groupdict()
                    loc = d["loc"]
                    if len(loc) == 4 and loc[0] == "K":
                        loc = loc[1:]
                    dist = int(d["dist"])
                    bearing = DRCT2DIR[d["dir"]]

                elif re.match(OV_LOCDIR, therest):
                    # KFAR330008
                    d = re.match(OV_LOCDIR, therest).groupdict()
                    loc = d["loc"]
                    if len(loc) == 4 and loc[0] == "K":
                        loc = loc[1:]
                    bearing = int(d["dir"])
                    dist = int(d["dist"])
                elif re.match(OV_LATLON, therest):
                    # 2500N07000W
                    # FMH-12 says this is in degrees and minutes!
                    d = re.match(OV_LATLON, therest).groupdict()
                    _pr.latitude = float(
                        "%s.%i"
                        % (
                            d["lat"][:-2],
                            int(float(d["lat"][-2:]) / 60.0 * 10000.0),
                        )
                    )
                    if d["latsign"] == "S":
                        _pr.latitude = 0 - _pr.latitude
                    _pr.longitude = float(
                        "%s.%i"
                        % (
                            d["lon"][:-2],
                            int(float(d["lon"][-2:]) / 60.0 * 10000.0),
                        )
                    )
                    if d["lonsign"] == "W":
                        _pr.longitude = 0 - _pr.longitude
                    continue
                elif therest == "O":
                    # Use the first part of the report in this case
                    loc = report[:3]
                elif therest.find("-") > 0 and re.match(OV_TWOLOC, therest):
                    d = re.match(OV_TWOLOC, therest).groupdict()
                    numbers = re.findall("[0-9]{6}", therest)
                    if numbers:
                        bearing = int(numbers[0][:3])
                        dist = int(numbers[0][3:])
                        loc = d["loc2"]
                        if len(loc) == 4 and loc[0] == "K":
                            loc = loc[1:]
                    else:
                        # Split the distance between the two points
                        lats = []
                        lons = []
                        for loc in [d["loc1"], d["loc2"]]:
                            if len(loc) == 4 and loc[0] == "K":
                                loc = loc[1:]
                            if loc not in self.nwsli_provider:
                                self.warnings.append(
                                    f"Unknown location: {loc} '{report}'"
                                )
                                return None
                            lats.append(self.nwsli_provider[loc]["lat"])
                            lons.append(self.nwsli_provider[loc]["lon"])
                        _pr.latitude = sum(lats) / 2.0
                        _pr.longitude = sum(lons) / 2.0
                        continue
                else:
                    loc = therest[:3]

                if loc not in self.nwsli_provider:
                    if _pr.base_loc is None:
                        self.warnings.append(
                            f"Unknown location: {loc} '{report}'"
                        )
                        return None
                    loc = _pr.base_loc
                    if loc not in self.nwsli_provider:
                        self.warnings.append(
                            f"Double-unknown location: {report}"
                        )
                        return None
                    # So we discard the offset when we go back to the base
                    dist = 0
                    bearing = 0
                _pr.longitude, _pr.latitude = self.compute_loc(
                    loc, dist, bearing
                )
                continue

            # Time
            if token.startswith("TM "):
                numbers = re.findall("[0-9]{4}", token)
                if len(numbers) != 1:
                    self.warnings.append("TM parse failed %s" % (report,))
                    return None
                hour = int(numbers[0][:2])
                minute = int(numbers[0][2:])
                _pr.valid = self.compute_pirep_valid(hour, minute)
                continue

        return _pr if _pr.latitude is not None else None

    def compute_loc(self, loc, dist, bearing):
        """ Figure out the lon/lat for this location """
        lat = self.nwsli_provider[loc]["lat"]
        lon = self.nwsli_provider[loc]["lon"]
        # shortcut
        if dist == 0:
            return lon, lat
        meters = distance(float(dist), "MI").value("M")
        northing = meters * math.cos(math.radians(bearing)) / 111111.0
        easting = (
            meters
            * math.sin(math.radians(bearing))
            / math.cos(math.radians(lat))
            / 111111.0
        )
        return lon + easting, lat + northing

    def compute_pirep_valid(self, hour, minute):
        """ Based on what utcnow is set to, compute when this is valid """
        res = self.utcnow.replace(
            hour=hour, minute=minute, second=0, microsecond=0
        )
        if hour > self.utcnow.hour:
            res -= datetime.timedelta(hours=24)
        return res

    def sql(self, txn):
        """ Save the reports to the database via the transaction """
        for report in self.reports:
            if report.is_duplicate:
                continue
            txn.execute(
                "INSERT into pireps(valid, geom, is_urgent, "
                "aircraft_type, report) VALUES (%s, "
                "ST_GeographyFromText('SRID=4326;POINT(%s %s)'),%s,%s,%s)",
                (
                    report.valid,
                    report.longitude,
                    report.latitude,
                    report.priority == Priority.UUA,
                    report.aircraft_type,
                    report.text,
                ),
            )

    def assign_cwsu(self, txn):
        """ Use this transaction object to assign CWSUs for the pireps """
        for report in self.reports:
            txn.execute(
                "select distinct id from cwsu WHERE "
                "st_contains(geom, geomFromEWKT('SRID=4326;POINT(%s %s)'))",
                (report.longitude, report.latitude),
            )
            if txn.rowcount == 0:
                # self.warnings.append("Find CWSU failed %.3f %.3f %s" % (
                #    report.longitude, report.latitude, report.text))
                continue
            row = txn.fetchone()
            report.cwsu = row["id"]

    def get_jabbers(self, _uri, _uri2=None):
        """ get jabber messages """
        res = []
        for report in self.reports:
            if report.is_duplicate or report.valid is None:
                continue
            jmsg = {
                "priority": "Urgent"
                if report.priority == Priority.UUA
                else "Routine",
                "ts": report.valid.strftime("%H%M"),
                "report": html_escape(report.text),
                "color": (
                    "#ff0000" if report.priority == Priority.UUA else "#00ff00"
                ),
            }
            plain = "%(priority)s pilot report at %(ts)sZ: %(report)s" % jmsg
            html = (
                "<span style='color:%(color)s;'>%(priority)s pilot "
                "report</span> at %(ts)sZ: %(report)s"
            ) % jmsg
            xtra = {
                "channels": (
                    f"{report.priority}.{report.cwsu},{report.priority}.PIREP"
                ),
                "geometry": "POINT(%s %s)"
                % (report.longitude, report.latitude),
                "ptype": report.priority,
                "category": "PIREP",
                "twitter": plain[:140],
                "valid": report.valid.strftime("%Y%m%dT%H:%M:00"),
            }
            res.append([plain, html, xtra])
        return res


def parser(buf, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """ A parser implementation """
    return Pirep(
        buf,
        utcnow=utcnow,
        ugc_provider=ugc_provider,
        nwsli_provider=nwsli_provider,
    )



"""See if we can do stuff with the network"""
# pylint: disable=redefined-outer-name

import psycopg2.extras
import pytest
from pyiem import network
from pyiem.util import get_dbconn


@pytest.fixture
def dbcursor():
    """With each test"""
    conn = get_dbconn("mesosite")
    cursor = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
    cursor.execute(
        """
        INSERT into stations(id, name, network, online)
        VALUES ('BOGUS', 'BOGUS NAME', 'BOGUS', 't')
        RETURNING iemid
    """
    )
    iemid = cursor.fetchone()[0]
    cursor.execute(
        """
        INSERT into station_attributes(iemid, attr, value)
        VALUES (%s, 'A', 'AA'), (%s, 'B', 'BB')
    """,
        (iemid, iemid),
    )
    cursor.execute(
        """
        INSERT into stations(id, name, network, online)
        VALUES ('BOGUS2', 'BOGUS2 NAME', 'BOGUS', 't')
    """
    )
    cursor.execute(
        """
        INSERT into stations(id, name, network, online)
        VALUES ('BOGUS3', 'BOGUS3 NAME', 'BOGUS2', 'f')
    """
    )
    return cursor


def test_basic(dbcursor):
    """ basic test of constructor """
    nt = network.Table("BOGUS", cursor=dbcursor)
    assert len(nt.sts.keys()) == 2

    nt = network.Table(["BOGUS", "BOGUS2"], cursor=dbcursor)
    assert len(nt.sts.keys()) == 2

    nt = network.Table(["BOGUS", "BOGUS2"], cursor=dbcursor, only_online=False)
    assert len(nt.sts.keys()) == 3

    assert nt.sts["BOGUS"]["name"] == "BOGUS NAME"
    assert len(nt.sts["BOGUS"]["attributes"]) == 2
    assert not nt.sts["BOGUS2"]["attributes"]

    # kind of a lame test here without the local cursor
    nt = network.Table(["BOGUS", "BOGUS2"])
    assert not nt.sts.keys()

    nt = network.Table("BOGUS2", cursor=dbcursor)
    assert not nt.sts.keys()



"""
Processing of GINI formatted data found on NOAAPORT
"""
import struct
import math
import zlib
from datetime import timezone, datetime
import os

import pyproj
import numpy as np
from pyiem.util import LOG

DATADIR = os.sep.join([os.path.dirname(__file__), "../data"])
M_PI_2 = 1.57079632679489661923
M_PI = 3.14159265358979323846
RE_METERS = 6371200.0
ENTITIES = [
    "UNK",
    "UNK",
    "MISC",
    "JERS",
    "ERS",
    "POES",
    "COMP",
    "DMSP",
    "GMS",
    "METEOSAT",
    "GOES7",
    "GOES8",
    "GOES9",
    "GOES10",
    "GOES11",
    "GOES12",
    "GOES13",
    "GOES14",
    "GOES15",
]
LABELS = [
    "UNK",
    "UNK",
    "MISC",
    "JERS",
    "ERS",
    "POES",
    "COMP",
    "DMSP",
    "GMS",
    "METEOSAT",
    "GOES",
    "GOES",
    "GOES",
    "GOES",
    "GOES",
    "GOES",
    "GOES",
    "GOES",
    "GOES",
]
CHANNELS = [
    "",
    "VIS",
    "3.9",
    "WV",
    "IR",
    "12",
    "13.3",
    "1.3",
    "U8",
    "U9",
    "U10",
    "U11",
    "U12",
    "LI",
    "PW",
    "SKIN",
    "CAPE",
    "TSURF",
    "WINDEX",
]
for _u in range(22, 100):
    CHANNELS.append(f"U{_u}")
SECTORS = [
    "NHCOMP",
    "EAST",
    "WEST",
    "AK",
    "AKNAT",
    "HI",
    "HINAT",
    "PR",
    "PRNAT",
    "SUPER",
    "NHCOMP",
    "CCONUS",
    "EFLOAT",
    "WFLOAT",
    "CFLOAT",
    "PFLOAT",
]

AWIPS_GRID_GUESS = {
    "A": 207,
    "B": 203,
    "E": 211,
    "F": 0,
    "H": 208,
    "I": 204,
    "N": 0,
    "P": 210,
    "Q": 205,
    "W": 211,
}

AWIPS_GRID = {
    "TIGB": 203,
    "TIGE": 211,
    "TIGW": 211,
    "TIGH": 208,
    "TIGP": 210,
    "TIGA": 207,
    "TIGI": 204,
    "TIGQ": 205,
    "TICF": 201,
}


def uint24(data):
    """convert three byte data that represents an unsigned int"""
    u = int(struct.unpack(">B", data[0:1])[0]) << 16
    u += int(struct.unpack(">B", data[1:2])[0]) << 8
    u += int(struct.unpack(">B", data[2:3])[0])
    return u


def int24(data):
    """Convert to int."""
    u = int(struct.unpack(">B", data[0:1])[0] & 127) << 16
    u += int(struct.unpack(">B", data[1:2])[0]) << 8
    u += int(struct.unpack(">B", data[2:3])[0])
    if (struct.unpack(">B", data[0:1])[0] & 128) != 0:
        u *= -1
    return u


def get_ir_ramp():
    """ Return a np 256x3 array of colors to use for IR """
    fn = "%s/gini_ir_ramp.txt" % (DATADIR,)
    data = np.zeros((256, 3), np.uint8)
    for i, line in enumerate(open(fn)):
        tokens = line.split()
        data[i, :] = [int(tokens[0]), int(tokens[1]), int(tokens[2])]
    return data


class GINIZFile:
    """
    Deal with compressed GINI files, which are the standard on NOAAPORT
    """

    def __init__(self, fobj):
        """Create a GNIFile instance with a compressed file object

        Args:
          fobj (file): A fileobject
        """
        fobj.seek(0)
        # WMO HEADER
        self.wmo = (fobj.read(21)).strip().decode("utf-8")
        d = zlib.decompressobj()
        hdata = d.decompress(fobj.read())
        self.metadata = self.read_header(hdata[21:])
        self.init_projection()
        totsz = len(d.unused_data)
        # 5120 value chunks, so we need to be careful!
        sdata = b""
        chunk = b"x\xda"
        i = 0
        for part in d.unused_data.split(b"x\xda"):
            if part == b"" and i == 0:
                continue
            chunk += part
            try:
                sdata += zlib.decompress(chunk)
                i += 1
                totsz -= len(chunk)
                chunk = b"x\xda"
            except Exception:
                chunk += b"x\xda"
        if totsz != 0:
            LOG.info("Totalsize left: %s", totsz)

        self.data = np.reshape(
            np.fromstring(sdata, np.int8),
            (self.metadata["numlines"] + 1, self.metadata["linesize"]),
        )

    def __str__(self):
        """return a string representation"""
        text = "%s Line Size: %s Num Lines: %s" % (
            self.wmo,
            self.metadata["linesize"],
            self.metadata["numlines"],
        )
        return text

    def awips_grid(self):
        """
        Return the awips grid number based on the WMO header
        """
        try1 = AWIPS_GRID.get(self.wmo[:4], None)
        if try1:
            return try1
        return AWIPS_GRID_GUESS.get(self.wmo[3], None)

    def current_filename(self):
        """
        Return a filename for this product, we'll use the format
        {SOURCE}_{SECTOR}_{CHANNEL}_{VALID}.png
        """
        return "%s_%s_%s.png" % (
            LABELS[self.metadata["creating_entity"]],
            SECTORS[self.metadata["sector"]],
            CHANNELS[self.metadata["channel"]],
        )

    def get_bird(self):
        """
        Return a string label for this satellite
        """
        return ENTITIES[self.metadata["creating_entity"]]

    def get_sector(self):
        """Return the sector."""
        return SECTORS[self.metadata["sector"]]

    def get_channel(self):
        """Return the channel."""
        return CHANNELS[self.metadata["channel"]]

    def archive_filename(self):
        """
        Return a filename for this product, we'll use the format
        {SOURCE}_{SECTOR}_{CHANNEL}_{VALID}.png
        """
        return ("%s_%s_%s_%s.png") % (
            LABELS[self.metadata["creating_entity"]],
            SECTORS[self.metadata["sector"]],
            CHANNELS[self.metadata["channel"]],
            self.metadata["valid"].strftime("%Y%m%d%H%M"),
        )

    def init_llc(self):
        """
        Initialize Lambert Conic Comformal
        """
        self.metadata["proj"] = pyproj.Proj(
            proj="lcc",
            lat_0=self.metadata["latin"],
            lat_1=self.metadata["latin"],
            lat_2=self.metadata["latin"],
            lon_0=self.metadata["lov"],
            a=6371200.0,
            b=6371200.0,
        )

        # s = 1.0
        # if self.metadata['proj_center_flag'] != 0:
        #    s = -1.0
        psi = M_PI_2 - abs(math.radians(self.metadata["latin"]))
        cos_psi = math.cos(psi)
        # r_E = RE_METERS / cos_psi
        alpha = math.pow(math.tan(psi / 2.0), cos_psi) / math.sin(psi)

        x0, y0 = self.metadata["proj"](
            self.metadata["lon1"], self.metadata["lat1"]
        )
        self.metadata["x0"] = x0
        self.metadata["y0"] = y0
        # self.metadata['dx'] *= alpha
        # self.metadata['dy'] *= alpha
        self.metadata["y1"] = y0 + (self.metadata["dy"] * self.metadata["ny"])

        (self.metadata["lon_ul"], self.metadata["lat_ul"]) = self.metadata[
            "proj"
        ](self.metadata["x0"], self.metadata["y1"], inverse=True)
        LOG.info(
            (
                "lat1: %.5f y0: %5.f y1: %.5f lat_ul: %.3f "
                "lat_ur: %.3f lon_ur: %.3f alpha: %.5f dy: %.3f"
            ),
            self.metadata["lat1"],
            y0,
            self.metadata["y1"],
            self.metadata["lat_ul"],
            self.metadata["lat_ur"],
            self.metadata["lon_ur"],
            alpha,
            self.metadata["dy"],
        )

    def init_mercator(self):
        """
        Compute mercator projection stuff
        """
        self.metadata["proj"] = pyproj.Proj(
            proj="merc",
            lat_ts=self.metadata["latin"],
            x_0=0,
            y_0=0,
            a=6371200.0,
            b=6371200.0,
        )
        x0, y0 = self.metadata["proj"](
            self.metadata["lon1"], self.metadata["lat1"]
        )
        self.metadata["x0"] = x0
        self.metadata["y0"] = y0

        x1, y1 = self.metadata["proj"](
            self.metadata["lon2"], self.metadata["lat2"]
        )
        self.metadata["x1"] = x1
        self.metadata["y1"] = y1

        self.metadata["dx"] = (x1 - x0) / self.metadata["nx"]
        self.metadata["dy"] = (y1 - y0) / self.metadata["ny"]

        (self.metadata["lon_ul"], self.metadata["lat_ul"]) = self.metadata[
            "proj"
        ](self.metadata["x0"], self.metadata["y1"], inverse=True)

        LOG.info(
            (
                "latin: %.2f lat_ul: %.3f lon_ul: %.3f "
                "y0: %5.f y1: %.5f dx: %.3f dy: %.3f"
            ),
            self.metadata["latin"],
            self.metadata["lat_ul"],
            self.metadata["lon_ul"],
            y0,
            y1,
            self.metadata["dx"],
            self.metadata["dy"],
        )

    def init_stereo(self):
        """
        Compute Polar Stereographic
        """
        self.metadata["proj"] = pyproj.Proj(
            proj="stere",
            lat_ts=60,
            lat_0=90,
            lon_0=self.metadata["lov"],
            x_0=0,
            y_0=0,
            a=6371200.0,
            b=6371200.0,
        )
        # First point!
        x0, y0 = self.metadata["proj"](
            self.metadata["lon1"], self.metadata["lat1"]
        )
        self.metadata["x0"] = x0
        self.metadata["y0"] = y0

        self.metadata["y1"] = y0 + (self.metadata["dy"] * self.metadata["ny"])
        (self.metadata["lon_ul"], self.metadata["lat_ul"]) = self.metadata[
            "proj"
        ](x0, self.metadata["y1"], inverse=True)

        LOG.info(
            (
                "lon_ul: %.2f lat_ul: %.2f "
                "lon_ll: %.2f lat_ll: %.2f "
                " lov: %.2f latin: %.2f lat1: %.2f lat2: %.2f "
                "y0: %5.f y1: %.5f dx: %.3f dy: %.3f"
            ),
            self.metadata["lon_ul"],
            self.metadata["lat_ul"],
            self.metadata["lon1"],
            self.metadata["lat1"],
            self.metadata["lov"],
            self.metadata["latin"],
            self.metadata["lat1"],
            self.metadata["lat2"],
            y0,
            self.metadata["y1"],
            self.metadata["dx"],
            self.metadata["dy"],
        )

    def init_projection(self):
        """
        Setup Grid and projection details
        """
        if self.metadata["map_projection"] == 3:
            self.init_llc()
        elif self.metadata["map_projection"] == 1:
            self.init_mercator()
        elif self.metadata["map_projection"] == 5:
            self.init_stereo()
        else:
            LOG.info("Unknown Projection: %s", self.metadata["map_projection"])

    def read_header(self, hdata):
        """read the header!"""
        meta = {}
        meta["source"] = struct.unpack("> B", hdata[0:1])[0]
        meta["creating_entity"] = struct.unpack("> B", hdata[1:2])[0]
        meta["sector"] = struct.unpack("> B", hdata[2:3])[0]
        meta["channel"] = struct.unpack("> B", hdata[3:4])[0]

        meta["numlines"] = struct.unpack(">H", hdata[4:6])[0]
        meta["linesize"] = struct.unpack(">H", hdata[6:8])[0]

        yr = 1900 + struct.unpack("> B", hdata[8:9])[0]
        mo = struct.unpack("> B", hdata[9:10])[0]
        dy = struct.unpack("> B", hdata[10:11])[0]
        hh = struct.unpack("> B", hdata[11:12])[0]
        mi = struct.unpack("> B", hdata[12:13])[0]
        ss = struct.unpack("> B", hdata[13:14])[0]
        # hs = struct.unpack("> B", hdata[14:15] )[0]
        meta["valid"] = datetime(yr, mo, dy, hh, mi, ss).replace(
            tzinfo=timezone.utc
        )
        meta["map_projection"] = struct.unpack("> B", hdata[15:16])[0]
        meta["proj_center_flag"] = struct.unpack("> B", hdata[36:37])[0] >> 7
        meta["scan_mode"] = struct.unpack("> B", hdata[37:38])[0]

        meta["nx"] = struct.unpack(">H", hdata[16:18])[0]
        meta["ny"] = struct.unpack(">H", hdata[18:20])[0]
        meta["res"] = struct.unpack(">B", hdata[41:42])[0]
        # Is Calibration Info included?
        # http://www.nws.noaa.gov/noaaport/document/ICD%20CH5-2005-1.pdf
        # page24
        # Mercator
        if meta["map_projection"] == 1:
            meta["lat1"] = int24(hdata[20:23])
            meta["lon1"] = int24(hdata[23:26])
            meta["lov"] = 0
            meta["dx"] = struct.unpack(">H", hdata[33:35])[0]
            meta["dy"] = struct.unpack(">H", hdata[35:37])[0]
            meta["latin"] = int24(hdata[38:41])
            meta["lat2"] = int24(hdata[27:30])
            meta["lon2"] = int24(hdata[30:33])
            meta["lat_ur"] = int24(hdata[55:58])
            meta["lon_ur"] = int24(hdata[58:61])
        # lambert == 3, polar == 5
        else:
            meta["lat1"] = int24(hdata[20:23])
            meta["lon1"] = int24(hdata[23:26])
            meta["lov"] = int24(hdata[27:30])
            meta["dx"] = uint24(hdata[30:33])
            meta["dy"] = uint24(hdata[33:36])
            meta["latin"] = int24(hdata[38:41])
            meta["lat2"] = 0
            meta["lon2"] = 0
            meta["lat_ur"] = int24(hdata[55:58])
            meta["lon_ur"] = int24(hdata[58:61])

        meta["dx"] = meta["dx"] / 10.0
        meta["dy"] = meta["dy"] / 10.0
        meta["lat1"] = meta["lat1"] / 10000.0
        meta["lon1"] = meta["lon1"] / 10000.0
        meta["lov"] = meta["lov"] / 10000.0
        meta["latin"] = meta["latin"] / 10000.0
        meta["lat2"] = meta["lat2"] / 10000.0
        meta["lon2"] = meta["lon2"] / 10000.0
        meta["lat_ur"] = meta["lat_ur"] / 10000.0
        meta["lon_ur"] = meta["lon_ur"] / 10000.0

        return meta
